/* 
 *   PenMount XFree86 Input driver
 *
 *      o PenMount 9000 (COM)
 *      o PenMount 5000 (USB)
 *      o PenMount 6000 (COM/USB)
 *
 *  2008/07/11 - from /dev/hidraw?
               - support Ubuntu 8.04, Fedora 9, Mandriva 2008.1, OpenSuSE 11
               - added some parameters
    2008/07/24 - add parameter 'ConfigFilePath'
    
 */
/* $XFree86: xc/programs/Xserver/hw/xfree86/input/penmount/xf86PM.c,v 1.2 2000/08/11 19:10:46 dawes Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define _PENMOUNT_C_

/*****************************************************************************
  #define PENMOUNT_XF86 : for XFree86 4.x
  #undef  PENMOUNT_XF86 : for Xorg
 ****************************************************************************/
#define    PENMOUNT_XF86 0


/*****************************************************************************
 *    Standard Headers
 ****************************************************************************/
#include <asm/types.h>
#include <linux/input.h>
#include <linux/hidraw.h>
#include <linux/kd.h>
#include <stdio.h>

/* keithp - a hack to avoid redefinitions of these in xf86str.h */
#ifdef BUS_PCI
#undef BUS_PCI
#endif
#ifdef BUS_ISA
#undef BUS_ISA
#endif

#include <misc.h>
#include <xf86.h>
#define NEED_XF86_TYPES
/*#include <xf86_ansic.h>*/
#include <xf86_libc.h>
#include <xf86_OSproc.h>
#include <xf86Xinput.h>
#include <xisb.h>
#include <exevents.h>

#include <sys/time.h>
#include <signal.h>
#include <errno.h>

/*#include <X11/keysym.h>*/

/*****************************************************************************
 *    Local Headers
 ****************************************************************************/
#include "xf86PM.h"
#include "calib.h"

#include "randrstr.h"


/*****************************************************************************
 *    Macros
 ****************************************************************************/
#define SYSCALL(call) while(((call) == -1) && (errno == EINTR))
#define ABS(x) ((x) > 0 ? (x) : -(x))
#define TRACE(lvl, f) {if ((lvl) == debug_level) f;}

/*#define IOCTL_COMMAND_BUFFER_LEN    10*/
#define IOCTL_COMMAND_BUFFER_LEN    5
#define IOCTL_SET_ADV_CALIBRATION   0xFF
#define IOCTL_SET_STD_CALIBRATION   0xFE


/*****************************************************************************
 *    Local Variables
 ****************************************************************************/
#ifndef PENMOUNT_XF86
_X_EXPORT InputDriverRec PENMOUNT =
#else
InputDriverRec PENMOUNT =
#endif
{
    1,                  /* driverVersion    */
    "penmount",         /* driverName       */
    NULL,               /* Identify         */
    PenMountPreInit,    /* PreInit          */
    PenMountUnInit,     /* UnInit           */
    NULL,               /* module           */
    0                   /* refCount         */
};

#ifdef XFree86LOADER
static XF86ModuleVersionInfo VersionRec =
{
    "penmount",             /* name of module                           */
    MODULEVENDORSTRING,     /* vendor specific string                   */
    MODINFOSTRING1,         
    MODINFOSTRING2,
#ifndef PENMOUNT_XF86
    XORG_VERSION_CURRENT,   /* contains XF86_VERSION_CURRENT            */
#else
    XF86_VERSION_CURRENT,   /* contains XF86_VERSION_CURRENT            */
#endif
    1, 3, 0,                /* major version, minor version, patchlevel */
    ABI_CLASS_XINPUT,       /* ABI class                                */
    ABI_XINPUT_VERSION,     /* ABI version                              */
    MOD_CLASS_XINPUT,       /* module class description                 */
    {0, 0, 0, 0}            /* checksum, digital signature of the version info structure */

};


static const char *reqSymbols[] = {
    "AddEnabledDevice",
    "ErrorF",
    "InitButtonClassDeviceStruct",
    "InitProximityClassDeviceStruct",
    "InitValuatorAxisStruct",
    "InitValuatorClassDeviceStruct",
    "InitPtrFeedbackClassDeviceStruct",
    "RemoveEnabledDevice",
    "Xcalloc",
    "Xfree",
    "XisbBlockDuration",
    "XisbFree",
    "XisbNew",
    "XisbRead",
    "XisbTrace",
    "screenInfo",
    "xf86AddInputDriver",
    "xf86AllocateInput",
    "xf86CloseSerial",
    "xf86CollectInputOptions",
    "xf86ErrorFVerb",
    "xf86FindOptionValue",
    "xf86GetMotionEvents",
    "xf86GetVerbosity",
    "xf86MotionHistoryAllocate",
    "xf86NameCmp",
    "xf86OpenSerial",
    "xf86CloseSerial",                /* second one ? */
    "xf86OptionListCreate",
    "xf86OptionListMerge",
    "xf86OptionListReport",
    "xf86PostButtonEvent",
    "xf86PostMotionEvent",
    "xf86PostProximityEvent",
    "xf86ProcessCommonOptions",
    "xf86RemoveLocalDevice",
    "xf86ScaleAxis",
    "xf86SetIntOption",
    "xf86SetStrOption",
    "xf86XInputSetScreen",
    "xf86XInputSetSendCoreEvents",
    NULL
};

/****************************************************************************************/

static pointer PenMountPlug( pointer module, pointer options, int *errmaj, int *errmin )
{
    /*xf86LoaderReqSymLists(reqSymbols, NULL);*/
    xf86AddInputDriver(&PENMOUNT, module, 0);

    /*return (pointer) 1;*/
    return module;
}

static void PenMountUnPlug(pointer p)
{
}


#ifndef PENMOUNT_XF86
_X_EXPORT XF86ModuleData penmountModuleData =
#else
XF86ModuleData penmountModuleData =
#endif
{
    &VersionRec,            /* vers     */
    PenMountPlug,           /* setup    */
    PenMountUnPlug          /* teardown */
};

#endif /* XFree86LOADER */


/* 
 * Be sure to set vmin appropriately for your device's protocol. You want to
 * read a full packet before returning
 */
static const char *default_options[] =
{
    "BaudRate",     "19200",
    "StopBits",     "1",
    "DataBits",     "8",
    "Parity",       "None",
    "Vmin",         "3",
    "Vtime",        "1",
    "FlowControl",  "None",
    NULL,
};

static int debug_level = 0;
static unsigned char g_last_packet[5];
static int bBrokenPipeCleanup = 1;
static int g_last_packet_x;
static int g_last_packet_y;
static int g_last_rotation;

#if 1
int createPenMountData(void)
{
    XF86FILE    *f;
    int     mode;
    
    mode = F_OK;
    
    if (xf86access("/tmp/penmount_calib", mode) != 0) {
        if (xf86access("/tmp/penmount_data", mode) != 0)
            xf86remove("/tmp/penmount_data");
        return -1;
    }

    f = xf86fopen("/tmp/penmount_data", "w");
    if (!f)
        return -1;  /* file not found */

    xf86fprintf(f, "%d ", g_last_packet_x);
    xf86fprintf(f, "%d ", g_last_packet_y);
    xf86fprintf(f, "%d ", g_last_rotation);
    xf86fclose(f);

}
#endif

/*****************************************************************************
 *    Function Definitions
 ****************************************************************************/
extern int     Calib_X[25];
extern int     Calib_Y[25];

void IOWorkerThread(InputInfoPtr pInfo)
{
    PenMountPrivatePtr priv = (PenMountPrivatePtr) (pInfo->private);
    int fd;
    int ret;
    

    if (xf86access("/tmp/penmount_calib_ok", F_OK) == 0) { /* exist */

        priv->touch_flags &= ~DOING_CALIB;

        xf86remove("/tmp/penmount_calib_ok");
        xf86remove("/tmp/penmount_calib");
        xf86remove("/tmp/penmount_data");

        ErrorF("%s IOWorkerThread Calib_XY[0] (before) : %d, %d\n", PM_INFO, Calib_X[0], Calib_Y[0]);
        ret = SetCalibration(priv->configfile, priv->rotation);
        priv->CalibrationMode = ret;
        ErrorF("%s IOWorkerThread Calibration Mode = %d\n", PM_INFO, ret);
        ErrorF("%s IOWorkerThread Calib_XY[0] (after ) : %d, %d\n", PM_INFO, Calib_X[0], Calib_Y[0]);

        priv->min_x = 0;
        priv->min_y = 0;
        priv->max_x = 2047;
        priv->max_y = 2047;

        /*
         * > 0, portraint mode, advanced calibration
         * = 0, standard calibration, reload calib data (pmode in the /etc/penmount.dat)
         * < 0, file not found, error
         */
        if (ret == 0)
            LoadStdCalibration(priv);

        /* restore hold time */
        /*
        priv->hold_time = priv->hold_time_orig;
        priv->hold_time_orig = 0;
        */

    }

    if (xf86access("/tmp/penmount_calib", F_OK) == 0) { /* exist */
        priv->touch_flags |= DOING_CALIB;
    }

}

static InputInfoPtr PenMountPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
{
    InputInfoPtr        pInfo;
    PenMountPrivatePtr  priv;
    char *s;
    char                model_string[96];
    FILE *f;
    int                 ret;
    char                devnode[32];
    int                 i;
    struct input_id     id;
    unsigned char       disable_6000_cmd[6] = { 0xf0, 0, 0, 0, 0, 0x0f };
    unsigned char       disable_9000_cmd[5] = { 0xf0, 0, 0, 0, 0 };
    int                 fd;

    priv = xcalloc (1, sizeof (PenMountPrivateRec));
    if (!priv)
        return NULL;

    pInfo = xf86AllocateInput(drv, 0);
    if (!pInfo) {
        xfree(priv);
        return NULL;
    }

    /* Process the options for your device like this */
    priv->pmode             = PM_DEF_PMODE;
    /* min_x, max_x, min_y, max_y are do for? */
    priv->min_x             = 0;
    priv->max_x             = 2047;
    priv->min_y             = 0;
    priv->max_y             = 2047;
    priv->screen_num        = 0;
    priv->screen_width      = -1;
    priv->screen_height     = -1;
    priv->swap_xy           = 0;
    priv->button_down       = FALSE;
    priv->button_number     = 1; /* why button_number = 2 in PenMount USB ? */
    priv->proximity         = FALSE;
    /*priv->chip              = CHIP_UNKNOWN;*/
    priv->pen_down          = 0;
    priv->CalibrationMode   = 0;
    priv->lex_mode          = PM_Byte0;


    pInfo->type_name        = XI_TOUCHSCREEN;

    /* enable/disable device */
    pInfo->device_control   = DeviceControl;

    /* It will be called when new data is available */
    pInfo->read_input       = ReadInput;

    pInfo->control_proc     = ControlProc;
    pInfo->close_proc       = CloseProc;

    /* toggle absolute/relative mode, it will be called
       when a client to swtich abs/rel mode
     */
    pInfo->switch_mode      = SwitchMode;

    pInfo->conversion_proc  = ConvertProc;
    pInfo->dev              = NULL;
    pInfo->private          = priv;
    pInfo->private_flags    = 0;
    pInfo->flags            = XI86_POINTER_CAPABLE | XI86_SEND_DRAG_EVENTS;
    pInfo->conf_idev        = dev;

    xf86CollectInputOptions(pInfo, default_options, NULL);

    xf86OptionListReport( pInfo->options );

    /* device node : /dev/hidraw? */
    priv->chip = CHIP_UNKNOWN;
    for (i=0; i<16; i++) {
        /* event0 ~ event16 */
        sprintf(devnode, "/dev/input/event%d", i);
        TRACE(2, ErrorF("%s Device Node : %s  ", PM_INFO, devnode));
        if ((fd = open(devnode, O_RDONLY)) < 0) {
            TRACE(2, ErrorF("%s open %s error!\n", PM_ERROR, devnode));
            continue;
        }
        ioctl(fd, EVIOCGID, &id);
        TRACE(2, ErrorF("%s BUS=%04x VID=%04x PID=%04x VER=%04x\n", PM_INFO,
                 id.bustype, id.vendor, id.product, id.version));
        if ((id.vendor == 0x14e1 && id.product == 0x6000) ||
            (id.vendor == 0x14e1 && id.product == 0x5000) ||
            (id.vendor == 0x1204 && id.product == 0x9998) ||
            (id.vendor == 0x14e1 && id.product == 0xfbfb)) {
            close(fd);
            sprintf(priv->evnode, "%s", devnode);
            ErrorF("%s PenMount %4X found! (%04x:%04x) @ %s\n", PM_INFO, id.product, id.vendor, id.product, devnode);
            if (id.product == 0x6000) {
                priv->chip = PM6000USB;
                sprintf(model_string, "SALT PenMount 6000 (USB) series");
                pInfo->read_input = ReadInput;
            } else if (id.product == 0x5000) {
                priv->chip = PM5000;
                sprintf(model_string, "SALT PenMount 5000 series");
                pInfo->read_input = ReadInput;
            } else if (id.product == 0x9998) {
                priv->chip = PM5000;
                sprintf(model_string, "SALT PenMount 5000 series (old)");
                pInfo->read_input = ReadInput;
            } else if (id.product == 0xfbfb) {
                priv->chip = PMLPC;
                sprintf(model_string, "Flybook PenMount LPC series");
                pInfo->read_input = ReadInputLpc;
            }

            break;
        } else {
            if (i==15)
                break;
            else
                continue;   /* continue to search */
        }
        close(fd);
    }

    if (priv->chip == CHIP_UNKNOWN) {
        s = xf86FindOptionValue (pInfo->options, "Protocol");
        priv->chip = PM9000;
        priv->packet_size = 5;
        pInfo->read_input = ReadInputComPort;
        if (s) {
            if (xf86NameCmp (s, "PM9000") == 0) {
                priv->chip = PM9000;
                sprintf(model_string, "SALT PenMount 9000 series");
            } else if (xf86NameCmp (s, "PM6000COM") == 0) {
                priv->chip = PM6000COM;
                priv->packet_size = 6;
                sprintf(model_string, "SALT PenMount 6000 (COM) series");
            }
        }
    }        

    if (priv->chip == PM9000 || priv->chip == PM6000COM) {
        pInfo->fd = xf86OpenSerial (pInfo->options);
        if (pInfo->fd == -1) {
            ErrorF("%s PenMount driver unable to open device.\n", PM_ERROR);
            goto SetupProc_fail;
        }
        xf86CloseSerial(pInfo->fd);
    }



    /* Process the options for your device like this */
    priv->screen_num    = xf86SetIntOption( pInfo->options, "ScreenNumber", 0 );
    
    /* sent which button at pen-down moment (1:left, 2:middle, 3:right) */
    priv->button_number = xf86SetIntOption( pInfo->options, "ButtonNumber", 1 );
    priv->swap_xy       = xf86SetIntOption( pInfo->options, "SwapXY", 0 );
    priv->buffer        = NULL;
    /* forced */
    /*priv->ad_bit        = 10;*/
    priv->ad_bit        = xf86SetIntOption( pInfo->options, "ADBit", 10 );

    /* 0:stream mode, 1:point mode */
    priv->pendown_mode  = xf86SetIntOption( pInfo->options, "PenDownMode", 0);
    ErrorF("%s PenDown Mode : %d\n", PM_INFO, priv->pendown_mode);


    /* Beep : 0=slience, 1=pen-down, 2=pen-up, 3=both */
    priv->beep        = xf86SetIntOption(pInfo->options, "Beep", 0);
    priv->press_vol   = xf86SetIntOption(pInfo->options, "PressVol", 100);
    priv->press_pitch = xf86SetIntOption(pInfo->options, "PressPitch", 880);
    priv->press_dur   = xf86SetIntOption(pInfo->options, "PressDur", 15) & 0xff;
    priv->rel_vol     = xf86SetIntOption(pInfo->options, "ReleaseVol", 0);
    priv->rel_pitch   = xf86SetIntOption(pInfo->options, "ReleasePitch", 1200);
    priv->rel_dur     = xf86SetIntOption(pInfo->options, "ReleaseDur", 10) & 0xff;
    priv->rightbutton = xf86SetIntOption(pInfo->options, "Button2" , 1);
    debug_level       = xf86SetIntOption(pInfo->options, "DebugLevel", 0);
    /* 500 ms */
    priv->hold_time   = xf86SetIntOption(pInfo->options, "HoldTime", 500000);
    /*priv->calib_hold_time = xf86SetIntOption(pInfo->options, "CalibHoldTime", 2000000);*/
    priv->calib_hold_time = xf86SetIntOption(pInfo->options, "CalibHoldTime", 1500000);
    priv->screen_scale = xf86SetIntOption(pInfo->options, "ScreenScale" , 0);
    priv->press_and_hold = xf86SetIntOption(pInfo->options, "PressAndHold" , 1);
    priv->lock_window_range = xf86SetIntOption(pInfo->options, "LockWindowRange" , 32);
    priv->xmin_offset = xf86SetIntOption(pInfo->options, "XMinOffset" , 10);
    priv->xmax_offset = xf86SetIntOption(pInfo->options, "XMaxOffset" , 10);
    priv->ymin_offset = xf86SetIntOption(pInfo->options, "YMinOffset" , 10);
    priv->ymax_offset = xf86SetIntOption(pInfo->options, "YMaxOffset" , 10);

    priv->std_xmin    = xf86SetIntOption(pInfo->options, "StdXMin" , 10);
    priv->std_xmax    = xf86SetIntOption(pInfo->options, "StdXMax" , 1000);
    priv->std_ymin    = xf86SetIntOption(pInfo->options, "StdYMin" , 10);
    priv->std_ymax    = xf86SetIntOption(pInfo->options, "StdYMax" , 1000);
    priv->average_point = xf86SetIntOption(pInfo->options, "AveragePoint" , 4);

    priv->rotation  = 1;
    
    s = xf86SetStrOption(pInfo->options, "ConfigFilePath", NULL);
    if (s)
        sprintf(priv->configfile, "%s%s", s, CALIBRATION_DATA_FILE);
    else
        sprintf(priv->configfile, "/etc%s", CALIBRATION_DATA_FILE);

    s = NULL;
    
    /* loading calibration data */
    ret = SetCalibration(priv->configfile, priv->rotation);
    ErrorF("%s Calibration Mode = %d\n", PM_INFO, ret);
    /* ret=0:std, ret<0:error ret>0:adv */
    if (ret <= 0) {
        priv->CalibrationMode = 0;
        LoadStdCalibration(priv);
    } else {
        priv->CalibrationMode = ret;
        /* DON'T CARE pmode in advanced calib */
        /*priv->pmode = ret;*/
        priv->min_x = 0;
        priv->min_y = 0;
        priv->max_x = 2047; /* forced to 10-bit */
        priv->max_y = 2047;
    }
  
    s = xf86FindOptionValue (pInfo->options, "ReportingMode");
    if ((s) && (xf86NameCmp (s, "raw") == 0))
        priv->reporting_mode = TS_Raw;
    else
        priv->reporting_mode = TS_Scaled;

    priv->proximity     = FALSE;
    priv->button_down   = FALSE;
    
    /* this results in an xstrdup that must be freed later */
    pInfo->name = xf86SetStrOption( pInfo->options, "DeviceName", model_string);
    xf86ProcessCommonOptions(pInfo, pInfo->options);

    pInfo->flags |= XI86_CONFIGURED;

    return (pInfo);

/* FAILED!!! FAILED!!! FAILED!!! FAILED!!! FAILED!!! FAILED!!! FAILED!!! */
SetupProc_fail:

    if ((pInfo) && (pInfo->fd))
        xf86CloseSerial(pInfo->fd);

    if ((pInfo) && (pInfo->name))
        xfree(pInfo->name);

    if ((priv) && (priv->buffer))
        XisbFree(priv->buffer);

    if (priv)
        xfree(priv);

    return (pInfo);
    /* why not return NULL? */
    /*return NULL;*/
}

static Bool DeviceInit (DeviceIntPtr dev)
{
    InputInfoPtr        pInfo = (InputInfoPtr) dev->public.devicePrivate;
    PenMountPrivatePtr  priv  = (PenMountPrivatePtr) (pInfo->private);
    unsigned char map[PM_MAX_BUTTONS];
    int i;
    
    for (i=0; i<PM_MAX_BUTTONS; i++)
        map[i] = i;

    /*
     * these have to be here instead of in the SetupProc, because when the
     * SetupProc is run at server startup, screenInfo is not setup yet
     */
    priv->screen_width  = screenInfo.screens[priv->screen_num]->width;
    priv->screen_height = screenInfo.screens[priv->screen_num]->height;

    /* Device reports button press for 1 button. */
     
    /* PenMount USB */
    if (InitButtonClassDeviceStruct (dev, PM_MAX_BUTTONS, map) == FALSE) {
        ErrorF("%s Unable to allocate PenMount ButtonClassDeviceStruct.\n", PM_ERROR);
        return !Success;
    }


    /*
     * Device reports motions on 2 axes in absolute coordinates.
     * Axes min and max values are reported in raw coordinates.
     *
     *   InitValuatorClassDeviceStruct(dev, numAxes, motionProc, numMotionEvents, mode)
     *
     *   GTK definitions
     *   axis 0 : x
     *   axis 1 : y
     *   axis 2 : pressure
     *   axis 3 : xtilt
     *   axis 4 : ytilt
     *   axis 5 : wheel
     */
/*    if (InitValuatorClassDeviceStruct (dev, 5, xf86GetMotionEvents,*/

    if (InitValuatorClassDeviceStruct (dev, 2, xf86GetMotionEvents,
        pInfo->history_size, Absolute) == FALSE) {
        ErrorF("%s Unable to allocate PenMount ValuatorClassDeviceStruct.\n", PM_ERROR);
        return !Success;
    } else {
    	/*InitValuatorAxisStruct(dev, 0, priv->min_x, priv->max_x,  9500, 0,  9500);*/
	    /* (dev, axes, min_val, max_val, resolution, min_resolution, max_resolution */
	    if (priv->chip == PM9000 || priv->chip == PM6000COM) {
        	InitValuatorAxisStruct(dev, 0, priv->min_x, priv->max_x, 9500, 0, 9500);
	        InitValuatorAxisStruct(dev, 1, priv->min_y, priv->max_y, 10500, 0, 10500);
	    } else {
        	InitValuatorAxisStruct(dev, 0, priv->min_x, priv->max_x, 10500, 0, 10500);
    	    InitValuatorAxisStruct(dev, 1, priv->min_y, priv->max_y, 10500, 0, 10500);
    	}
	    /*
    	InitValuatorAxisStruct(dev, 0, 0, 2047, 10500, 0, 10500);
	    InitValuatorAxisStruct(dev, 1, 0, 2047, 10500, 0, 10500);
	    */

    }

    if (InitProximityClassDeviceStruct(dev) == FALSE) {
        ErrorF("%s unable to allocate PenMount ProximityClassDeviceStruct.\n", PM_ERROR);
        return !Success;
    }
        
    if (InitPtrFeedbackClassDeviceStruct(dev, PenMountPtrCtrl) == FALSE) {
        ErrorF("%s unable to allocate PenMount PtrFeedbackClassDeviceStruct\n", PM_ERROR);
        return !Success;
    }

    /* Allocate the motion events buffer. */
    xf86MotionHistoryAllocate (pInfo);

    return (Success);

}


static Bool DeviceOn (DeviceIntPtr dev)
{
    InputInfoPtr        pInfo = (InputInfoPtr) dev->public.devicePrivate;
    PenMountPrivatePtr  priv  = (PenMountPrivatePtr) (pInfo->private);
    int                 fd = -1;
    struct hidraw_devinfo dinfo;
    struct input_id     id;
    char                devnode[32];
    int                 i;
    unsigned char       enable_6000_cmd[6] = { 0xf1, 0, 0, 0, 0, 0x0e };
    unsigned char       enable_9000_cmd[5] = { 0xf1, 0, 0, 0, 0 };

/**********************************************************************/
    if (priv->chip == PM6000USB || priv->chip == PM5000) {

        /*  /dev/input/eventX  */
        priv->evnode[0] = NULL;
        xf86memset(priv->evnode, 0, sizeof(priv->evnode));
        for (i=0; i<16; i++) {
            sprintf(devnode, "/dev/input/event%d", i);
            TRACE(2, ErrorF("%s Device Node : %s  ", PM_INFO, devnode));
            if ((fd = open(devnode, O_RDONLY)) < 0) {
                TRACE(2, ErrorF("%s open %s error!\n", PM_INFO, devnode));
                continue;
            }
            ioctl(fd, EVIOCGID, &id);
            TRACE(2, ErrorF("%s BUS=%04x VID=%04x PID=%04x VER=%04x\n", PM_INFO,
                     id.bustype, id.vendor, id.product, id.version));
            if ((id.vendor == 0x14e1 && id.product == 0x6000) ||
                (id.vendor == 0x14e1 && id.product == 0x5000) ||
                (id.vendor == 0x1204 && id.product == 0x9998)) {
                    
                close(fd);
                sprintf(priv->evnode, "%s", devnode);   /* remember event node */
                TRACE(2, ErrorF("%s Device Node : %s Found!\n", PM_INFO, devnode));
                ErrorF("%s PenMount %4X found! (%04x:%04x) @ %s\n", PM_INFO, id.product, id.vendor, id.product, devnode);
                break;
            } else {
                if (i==15) {
                    ErrorF("%s Can't find PenMount!\n", PM_ERROR);
                    close(fd);
                    break;
                } else {
                    continue;   /* continue to search */
                }
            }
            close(fd);
        }

        /*  /dev/hidrawX  */
        for (i=0; i<8; i++) {
            sprintf(devnode, "/dev/hidraw%d", i);
            TRACE(2, ErrorF("%s Device Node : %s  ", PM_INFO, devnode));

            if ((fd = open(devnode, O_RDONLY)) < 0) {
                TRACE(2, ErrorF("%s open %s error!\n", PM_INFO, devnode));
                continue;
            }
            
            ioctl(fd, HIDIOCGRAWINFO, &dinfo);
            TRACE(2, ErrorF("%s VID=%04x  PID=%04x\n", PM_INFO, dinfo.vendor, dinfo.product));

            if ( (dinfo.vendor == 0x14e1 && dinfo.product == 0x6000) ||
                 (dinfo.vendor == 0x14e1 && dinfo.product == 0x5000) ||
                 (dinfo.vendor == 0x1204 && dinfo.product == 0x9998)) {
                close(fd);
                pInfo->fd = open(devnode, O_RDWR);
                TRACE(2, ErrorF("%s Device Node : %s Found!\n", PM_INFO, devnode));
                ErrorF("%s PenMount %4X found! (%04x:%04x) @ %s\n",
                        PM_INFO, dinfo.product, dinfo.vendor, dinfo.product, devnode);
                break;
            } else {
                if (i==7) { /* cannot find /dev/hidraw? */
                    break;
                } else {
                    continue;   /* continue to search */
                }
            }
            close(fd);
            return !Success;        
        }

        /* grab /dev/input/eventX */
        /*if (priv->evnode[0] != NULL) {*/
        if (*(priv->evnode)) {
            priv->evfd = -1;
            if ((priv->evfd = open(priv->evnode, O_RDONLY)) > 0) {
                i = 0xffff;
                if (ioctl(priv->evfd, EVIOCGRAB, &i)) {
                    TRACE(2, ErrorF("%s evdev ioctl (EVIOCGRAB) error!\n", PM_ERROR, priv->evnode));
                }
            } else {
                    TRACE(2, ErrorF("%s open %s error!\n", PM_ERROR, priv->evnode));
            }
        }

        /*pInfo->fd = xf86OpenSerial(pInfo->options);*/
        if (pInfo->fd == -1) {
            xf86Msg(X_WARNING, "%s: cannot open input device\n", pInfo->name);
            close(priv->evfd);
            return (!Success);
        } else {
            priv->buffer = XisbNew(pInfo->fd, PM_BUFFER_SIZE);
            if (!priv->buffer) {
                xf86CloseSerial(pInfo->fd);
                pInfo->fd = -1;
                close(priv->evfd);
                return (!Success);
            }
        }

        XisbBlockDuration (priv->buffer, -1);
    
        /* 2008.05.13 added */
        /*xf86FlushInput(pInfo->fd);*/

    } else if (priv->chip == PMLPC) {  /* PMLPC */

        priv->evnode[0] = NULL;
        xf86memset(priv->evnode, 0, sizeof(priv->evnode));
        /* /dev/input/eventXX */
        for (i=0; i<16; i++) {
            sprintf(devnode, "/dev/input/event%d", i);
            TRACE(2, ErrorF("%s Device Node : %s  ", PM_INFO, devnode));
            if ((fd = open(devnode, O_RDONLY)) < 0) {
                TRACE(2, ErrorF("%s open %s error!\n", PM_INFO, devnode));
                continue;
            }
            ioctl(fd, EVIOCGID, &id);
            TRACE(2, ErrorF("%s BUS=%04x VID=%04x PID=%04x VER=%04x\n", PM_INFO,
                     id.bustype, id.vendor, id.product, id.version));
            if (id.vendor == 0x14e1 && id.product == 0xfbfb) {
                close(fd);
                sprintf(priv->evnode, "%s", devnode);
                pInfo->fd = open(devnode, O_RDONLY);
                /* grab /dev/input/eventXX */
                if (ioctl(pInfo->fd, EVIOCGRAB, (void*)1)) {
                    TRACE(2, ErrorF("%s evdev ioctl (EVIOCGRAB) error!\n", PM_ERROR, priv->evnode));
                }

                TRACE(2, ErrorF("%s Device Node : %s Found!\n", PM_INFO, devnode));
                ErrorF("%s PenMount %4X found! (%04x:%04x) @ %s\n", PM_INFO, id.product, id.vendor, id.product, devnode);
                break;
            } else {
                if (i==15) {
                    ErrorF("%s Can't find PenMount LPC!\n", PM_ERROR);
                    close(fd);
                    break;
                } else {
                    continue;   /* continue to search */
                }
            }
            close(fd);
        }

        if (pInfo->fd == -1) {
            xf86Msg(X_WARNING, "%s: cannot open input device\n", pInfo->name);
            return (!Success);
        } else {
            priv->buffer = XisbNew(pInfo->fd, PM_BUFFER_SIZE);
            if (!priv->buffer) {
                xf86CloseSerial(pInfo->fd);
                pInfo->fd = -1;
                return (!Success);
            }
        }
        XisbBlockDuration (priv->buffer, -1);
        
        priv->lex_mode = 0;

    } else if (priv->chip == PM6000COM || priv->chip == PM9000) {  /* PM9000, PM6000COM */

        pInfo->fd = xf86OpenSerial(pInfo->options);
        if (pInfo->fd == -1) {
            xf86Msg(X_WARNING, "%s: cannot open input device\n", pInfo->name);
            return (!Success);
        } else {
            priv->buffer = XisbNew(pInfo->fd, PM_BUFFER_SIZE);
            if (!priv->buffer) {
                xf86CloseSerial(pInfo->fd);
                pInfo->fd = -1;
                return (!Success);
            }
        }

#if 0
        XisbBlockDuration (priv->buffer, 500000);
        /* enable PenMount */
        if (QueryHardware(pInfo) != Success) {
            ErrorF("%s cannot find PenMount device", PM_ERROR);
            return (!Success);
        }
#endif

        /*xf86FlushInput(pInfo->fd);*/

        switch (priv->chip) {
            case PM6000COM:
                XisbWrite(priv->buffer, enable_6000_cmd, 6);
                break;
            case PM9000:
                XisbBlockDuration (priv->buffer, 500000);
                XisbWrite(priv->buffer, enable_9000_cmd, 5);
                PenMountGetPacket(priv);
                /* set range to mininum */
                enable_9000_cmd[0] = 0xf3;
                XisbWrite(priv->buffer, enable_9000_cmd, 5);
                PenMountGetPacket(priv);
                break;
            default:
                break;
        }

        XisbBlockDuration (priv->buffer, -1);

        priv->lex_mode = PM_Byte0;

        xf86FlushInput(pInfo->fd);
    }

    AddEnabledDevice (pInfo->fd);
    
    dev->public.on = TRUE;


    return (Success);
}

static Bool DeviceOff (DeviceIntPtr dev)
{
    InputInfoPtr        pInfo = (InputInfoPtr) dev->public.devicePrivate;
    PenMountPrivatePtr  priv  = (PenMountPrivatePtr) (pInfo->private);

    if (pInfo->fd != -1) {
        ioctl(priv->evfd, EVIOCGRAB, (void*)0);

        RemoveEnabledDevice (pInfo->fd);
        if (priv->buffer) {
            XisbFree(priv->buffer);
            priv->buffer = NULL;
        }
        xf86CloseSerial(pInfo->fd);
    }
    dev->public.on = FALSE;

    if (priv->chip == PM6000USB || priv->chip == PM5000)
        close(priv->evfd);

    return (Success);
}

static Bool DeviceControl (DeviceIntPtr dev, int mode)
{
    Bool    result;

    /* debug */
    TRACE(1, ErrorF("%s DeviceControl\n", PM_INFO));


    switch (mode) {
        case DEVICE_INIT:
            result = DeviceInit(dev);
            break;
        case DEVICE_ON:
            result = DeviceOn(dev);
            break;
        case DEVICE_OFF:
        case DEVICE_CLOSE:
            result = DeviceOff(dev);
            break;
        default:
            result = BadValue;
    }

    return result;
}


/* 
 * The ReadInput function will have to be tailored to your device
 */
#define GET_TIME(tv) (gettimeofday(&tv, (struct timezone *)NULL))
#define DIF_TIME(t1,t2) ((t2.tv_sec -t1.tv_sec) *1000000+ \
                           (t2.tv_usec-t1.tv_usec))
                           
static struct timeval tv0, tv1; /* time of last up, and down events */

static void ReadInput (InputInfoPtr pInfo)
{
    PenMountPrivatePtr priv = (PenMountPrivatePtr) (pInfo->private);
    int x, y;
    int button = 0;
    int temp = 0;
    int pmode = 1;

    char                buf[200];
    char                *pbuf;
    ssize_t             len;
    int                 ax, ay;
    ScrnInfoPtr pScrn;
    Rotation rotation;

    TRACE(1, ErrorF("%s ReadInput\n", PM_INFO));

    /* Get screen resolution again, beacase Ubuntu will change resolution after login X window */
    priv->screen_width  = screenInfo.screens[priv->screen_num]->width;
    priv->screen_height = screenInfo.screens[priv->screen_num]->height;

    TRACE(3, ErrorF("%s screen %d x %d\n", PM_INFO, priv->screen_width, priv->screen_height));

    /* unused, because of USB HID packet */
    /*priv->packet_size = 6;*/

    x = priv->old_x;
    y = priv->old_y;
    /*button = priv->old_button;*/

    /* 
     * set blocking to -1 on the first call because we know there is data to
     * read. Xisb automatically clears it after one successful read so that
     * succeeding reads are preceeded buy a select with a 0 timeout to prevent
     * read from blocking indefinately.
     */
    XisbBlockDuration (priv->buffer, -1);

    /* We use SYSCALL to make sure our call isn't interrupted */
    SYSCALL(len = read(pInfo->fd, buf, 200));  /* 200 = buffer size*/

    TRACE(2, ErrorF("%s ReadInput: len = %d\n", PM_INFO, len));

    if (len <= 0) {
        ErrorF("%s Error reading PenMount USB device : %s\n", PM_INFO, stderror(errno));
        return;
    }
    
    for (pbuf = (char*)buf; len > 0; ) {
        if (*pbuf== 0x30 || *pbuf==0x70) {
            button = (*pbuf++ & 0xff);
            temp   = (*pbuf++ & 0xff);
            x      = (*pbuf++ & 0xff);
            x <<= 8;
            x += temp;
            temp   = (*pbuf++ & 0xff);
            y      = (*pbuf++ & 0xff);
            y <<= 8;
            y += temp;
            len -= 5;
            TRACE(2, ErrorF("%s (%d, %d), button=%02x\n", PM_INFO, x, y, button));

            if (priv->ad_bit == 12) {
                x >>= 1;    /* 11-bit */
                y >>= 1;
            } else {
                x <<= 1;    /* 11-bit */
                y <<= 1;
            }
            
            priv->old_pendown = priv->pen_down;
            if (button == 0x70)
                priv->pen_down = 1;
            else
                priv->pen_down = 0;
        } else {
            pbuf++;
            len--;
            continue;
        }
        

        if (priv->old_pendown == 0 && priv->pen_down == 1) {    /* get pen-down moment */
            GET_TIME(tv0);
            
            priv->touch_flags = 0;  /* clear touch_flags */
            
            /* hold X, Y (raw) */
            priv->hold_x = x;
            priv->hold_y = y;

            priv->old_x  = x;
            priv->old_y  = y;
            priv->touch_flags |= LOCK_WINDOW;
            priv->touch_flags |= PEN_DOWN;
            
            PenMountBeep(priv, 1);
            
            priv->mcnt = 0;

            /* FIFO operation */
            IOWorkerThread(pInfo);

            g_last_packet_x = x;
            g_last_packet_y = y;

            TRACE(2, ErrorF("%s PEN-DOWN!!\n", PM_INFO));

            /******************  detect ROTATION *************/
            pScrn = xf86Screens[priv->screen_num];
            rotation = RRGetRotation(pScrn->pScreen);
            TRACE(2, ErrorF("PenMount : rotation = %d \n", rotation));
            if (rotation != priv->rotation) {
                priv->rotation = rotation;
                priv->CalibrationMode = SetCalibration(priv->configfile, priv->rotation);
                if (priv->CalibrationMode == 0)
                    LoadStdCalibration(priv);
            }
            g_last_rotation = rotation;

        } else if (priv->old_pendown == 1 && priv->pen_down == 0) { /* pen-up moment */
            priv->touch_flags |= PEN_UP;
            PenMountBeep(priv, 0);
            TRACE(2, ErrorF("%s PEN-UP!!\n", PM_INFO));
        }

        priv->old_x = x;
        priv->old_y = y;

        if (priv->CalibrationMode > 0) {
            /*************************************
             *       advanced calibration        *
             *************************************/
            TRACE(4, ErrorF("%s do_calib: raw(%d, %d)-->", PM_INFO, x, y));
            /* DO IT! */
            DoCalibrate(priv->xmin_offset, priv->xmax_offset, priv->ymin_offset, priv->ymax_offset, &x, &y);
            TRACE(4, ErrorF("%s do_advcal(%d, %d)-->", PM_INFO, x, y));

            if (priv->screen_scale == 1) {
                /* DON'T SCALED X, Y PLEASE..... */
                /*
                x = (x) * priv->screen_width  / (1023);
                y = (y) * priv->screen_height / (1023);
                */
                x = (x) * priv->screen_width  / (2047);
                y = (y) * priv->screen_height / (2047);
            }
            TRACE(4, ErrorF("%s scr(%d, %d), %dx%d\n", PM_INFO, x, y, priv->screen_width, priv->screen_height));
            
        } else {
            /*************************************
             *       standard calibration        *
             *************************************/

            /* Scaling */
            TRACE(4, ErrorF("%s do_stdcal: raw(%d, %d)-->", PM_INFO, x, y));

            if (priv->reporting_mode == TS_Scaled) {
                pmode = priv->pmode;
                
                /* swap x, y */
                if (pmode >= 5 && pmode <= 8) {
                    temp = x;
                    x    = y;
                    y    = temp;
                }

                if (pmode == 1 || pmode == 2 || pmode == 5 || pmode == 6) {
                    if (x < priv->std_xmin)
                        x = priv->std_xmin;
                    if (x > priv->std_xmax)
                        x = priv->std_xmax;
                    if (priv->screen_scale == 1)
                        x = ((x - priv->std_xmin) * priv->screen_width) / (priv->std_xmax - priv->std_xmin);
                    else
                        /*x = ((x - priv->std_xmin) * 1023) / (priv->std_xmax - priv->std_xmin);*/
                        x = ((x - priv->std_xmin) * 2047) / (priv->std_xmax - priv->std_xmin);
                } else {
                    if (x < priv->std_xmin)
                        x = priv->std_xmin;
                    if (x > priv->std_xmax)
                        x = priv->std_xmax;
                    if (priv->screen_scale == 1)
                        x = ((priv->std_xmax - x) * priv->screen_width) / (priv->std_xmax - priv->std_xmin);
                    else
                        /*x = ((priv->std_xmax - x) * 1023) / (priv->std_xmax - priv->std_xmin);*/
                        x = ((priv->std_xmax - x) * 2047) / (priv->std_xmax - priv->std_xmin);
                }

                if (pmode == 1 || pmode == 3 || pmode == 5 || pmode == 7) {
                    if (y < priv->std_ymin)
                        y = priv->std_ymin;
                    if (y > priv->std_ymax)
                        y = priv->std_ymax;
                    if (priv->screen_scale == 1)
                        y = ((y - priv->std_ymin) * priv->screen_height) / (priv->std_ymax - priv->std_ymin);
                    else
                        /*y = ((y - priv->std_ymin) * 1023) / (priv->std_ymax - priv->std_ymin);*/
                        y = ((y - priv->std_ymin) * 2047) / (priv->std_ymax - priv->std_ymin);
                } else {
                    if (y < priv->std_ymin)
                        y = priv->std_ymin;
                    if (y > priv->std_ymax)
                        y = priv->std_ymax;
                    if (priv->screen_scale == 1)
                        y = ((priv->std_ymax - y) * priv->screen_height) / (priv->std_ymax - priv->std_ymin);
                    else
                        /*y = ((priv->std_ymax - y) * 1023) / (priv->std_ymax - priv->std_ymin);*/
                        y = ((priv->std_ymax - y) * 2047) / (priv->std_ymax - priv->std_ymin);
                }
                TRACE(4, ErrorF("%s scr(%d, %d)\n", PM_INFO, x, y));
            }
        }/*end-if (CalibrationMode > 0)*/

        xf86XInputSetScreen(pInfo, priv->screen_num, x, y);

        /* 2008.05.13 added */
        if ((priv->proximity == FALSE) && (priv->pen_down == 1)) {
            priv->proximity = TRUE;
            xf86PostProximityEvent (pInfo->dev, 1, 0, 2, x, y);
        }

        /*
         * Send event
         * We *must* generate a motion before a button change if pointer
         * location has changed as DIX assumes this. This is why we always
         * emit a motion, regardless of the kind of packet processed.
         */
        switch(priv->average_point) {
            case 4:
                priv->av_x[priv->mcnt%4] = x;
                priv->av_y[priv->mcnt%4] = y;
                if (priv->mcnt < 4) {
                    ax = x;
                    ay = y;
                    xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, x, y);
                    TRACE(2, ErrorF("%s Motion[%d] : (%d, %d)\n", PM_INFO, priv->mcnt, x, y));
                } else {
                    ax = (priv->av_x[0]+priv->av_x[1]+priv->av_x[2]+priv->av_x[3])/4;
                    ay = (priv->av_y[0]+priv->av_y[1]+priv->av_y[2]+priv->av_y[3])/4;
                    /* slow down for PenMount 6000 USB */
                    if (priv->chip == PM6000USB) {
                        if (priv->mcnt%2)
                            xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, ax, ay);
                    } else {
                        xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, ax, ay);
                    }
                }
                break;
            case 8:
                priv->av_x[priv->mcnt%8] = x;
                priv->av_y[priv->mcnt%8] = y;
                if (priv->mcnt < 8) {
                    ax = x;
                    ay = y;
                    xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, x, y);
                    TRACE(2, ErrorF("%s Motion[%d] : (%d, %d)\n", PM_INFO, priv->mcnt, x, y));
                } else {
                    ax = (priv->av_x[0]+priv->av_x[1]+priv->av_x[2]+priv->av_x[3]+
                          priv->av_x[4]+priv->av_x[5]+priv->av_x[6]+priv->av_x[7])/8;
                    ay = (priv->av_y[0]+priv->av_y[1]+priv->av_y[2]+priv->av_y[3]+
                          priv->av_y[4]+priv->av_y[5]+priv->av_y[6]+priv->av_y[7])/8;
                    /* slow down for PenMount 6000 USB */
                    if (priv->chip == PM6000USB) {
                        if (priv->mcnt%2)
                            xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, ax, ay);
                    } else {
                        xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, ax, ay);
                    }
                }
                break;
            default:
                /*xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, x, y);*/
                ax = x;
                ay = y;
                xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, ax, ay);
                break;
        }/* switch */
        priv->mcnt++;


        /********************/
        /*  Press and Hold  */
        /********************/
        if (priv->touch_flags & LOCK_WINDOW) {
            GET_TIME(tv1);

            temp = g_last_packet_x + priv->old_x;
            g_last_packet_x = temp / 2;

            temp = g_last_packet_y + priv->old_y;
            g_last_packet_y = temp / 2;


            if (priv->touch_flags & DOING_CALIB) {
                /* calib hold time > 2000ms */
                if (DIF_TIME(tv0, tv1) > priv->calib_hold_time) {
                    priv->touch_flags |= ISSUE_2ND_BUTTON;
                    priv->touch_flags &= ~LOCK_WINDOW;
                }
                /* exceed range */
                if (xf86abs(priv->hold_x - priv->old_x) >= priv->lock_window_range ||
                    xf86abs(priv->hold_y - priv->old_y) >= priv->lock_window_range)
                    priv->touch_flags &= ~LOCK_WINDOW;
            } else {
                /* hold time > 700ms */
                if (DIF_TIME(tv0, tv1) > priv->hold_time) {
                    priv->touch_flags |= ISSUE_2ND_BUTTON;
                    priv->touch_flags &= ~LOCK_WINDOW;
                }
                /* exceed range */
                if (xf86abs(priv->hold_x - priv->old_x) >= priv->lock_window_range ||
                    xf86abs(priv->hold_y - priv->old_y) >= priv->lock_window_range)
                    priv->touch_flags &= ~LOCK_WINDOW;
            }
        }

        /* xf86PostButtonEvent(device, is_absolute, button, is_down, first_valuator, num_valuator, valuator list....)
         * button : (1=left, 3=right)
         * is_down: (1=down, 0=up) */


        /****************/
        /*  2nd Button  */
        /****************/
        if (priv->touch_flags & ISSUE_2ND_BUTTON) {
            priv->touch_flags &= ~(ISSUE_2ND_BUTTON);

            createPenMountData();   /* prior to send post button event */

            /* Send Left released before send RIGHT/MIDDLE pressed */
            xf86PostButtonEvent (pInfo->dev, TRUE, 1, 0, 0, 2, ax, ay);   /* LEFT released */
            priv->touch_flags &= ~BUTTON_LEFT_PRESSED;

            if (priv->rightbutton == 1) {
                priv->touch_flags |= BUTTON_RIGHT_PRESSED;
                xf86PostButtonEvent (pInfo->dev, TRUE, 3, 1, 0, 2, ax, ay);   /* RIGHT pressed */
                TRACE(2, ErrorF("%s (%d, %d)  RIGHT pressed\n", PM_INFO, ax, ay));
            } else {
                priv->touch_flags |= BUTTON_MIDDLE_PRESSED;
                xf86PostButtonEvent (pInfo->dev, TRUE, 2, 1, 0, 2, ax, ay);   /* MIDDLE pressed */
                TRACE(2, ErrorF("%s (%d, %d)  MIDDLE pressed\n", PM_INFO, ax, ay));
            }

        }


        /*******************/
        /*    Pen Down     */
        /*******************/
        if (priv->touch_flags & PEN_DOWN) {
            priv->touch_flags &= ~PEN_DOWN;
            /* FIFO operation */
            /*IOWorkerThread(pInfo);*/
            /*createPenMountData();*/   /* prior to send post button event */

            #if 0
            if (xf86access("/tmp/penmount_calib", F_OK) == 0) {
                priv->hold_time_orig = priv->hold_time;
                priv->hold_time = 2000000;  /* 2 sec */
            }
            #endif


            if (priv->pendown_mode == 0) {
                /* stream mode */
                priv->touch_flags |= BUTTON_LEFT_PRESSED;
                xf86PostButtonEvent (pInfo->dev, TRUE, 1, 1, 0, 2, ax, ay);   /* LEFT pressed */
                TRACE(2, ErrorF("%s (%d, %d)  LEFT pressed\n", PM_INFO, ax, ay));
            } else {
                /* point mode */
                xf86PostButtonEvent (pInfo->dev, TRUE, 1, 1, 0, 2, ax, ay);   /* LEFT pressed */
                xf86PostButtonEvent (pInfo->dev, TRUE, 1, 0, 0, 2, ax, ay);   /* LEFT released*/
                TRACE(2, ErrorF("%s (%d, %d)  LEFT pressed-release\n", PM_INFO, ax, ay));
            }
            /* save pendown X, Y (screen) for something */
            priv->pd_x = x;
            priv->pd_y = y;
        }


        /******************/
        /*     Pen Up     */
        /******************/
        if (priv->touch_flags & PEN_UP) {
            /*createPenMountData();*/   /* prior to send post button event */
            priv->touch_flags &= ~PEN_UP;

            if (priv->touch_flags & BUTTON_LEFT_PRESSED) {
                /* LEFT pressed */
                priv->touch_flags &= ~BUTTON_LEFT_PRESSED;
                if (priv->touch_flags & LOCK_WINDOW) {
                    /* LEFT released*/
                    /*
                    xf86PostButtonEvent (pInfo->dev, TRUE, 1, 0, 0, 2, priv->pd_x, priv->pd_y);
                    TRACE(2, ErrorF("(%d, %d)  LEFT released (LOCK_WINDOW)\n", priv->pd_x, priv->pd_y));
                    */
                    xf86PostButtonEvent (pInfo->dev, TRUE, 1, 0, 0, 2, ax, ay);
                    TRACE(2, ErrorF("%s (%d, %d)  LEFT released (LOCK_WINDOW)\n", PM_INFO, ax, ay));

                } else {
                    xf86PostButtonEvent (pInfo->dev, TRUE, 1, 0, 0, 2, ax, ay);   /* LEFT released*/
                    TRACE(2, ErrorF("%s (%d, %d)  LEFT released (none LOCK_WINDOW)\n", PM_INFO, ax, ay));
                }
            }

            if (priv->touch_flags & BUTTON_RIGHT_PRESSED) {
                /* RIGHT pressed */
                priv->touch_flags &= ~BUTTON_RIGHT_PRESSED;
                xf86PostButtonEvent (pInfo->dev, TRUE, 3, 0, 0, 2, ax, ay);   /* RIGHT released */
                TRACE(2, ErrorF("%s (%d, %d)  RIGHT released\n", PM_INFO, ax, ay));
            } else if (priv->touch_flags & BUTTON_MIDDLE_PRESSED) {
                /* MIDDLE pressed */
                priv->touch_flags &= ~BUTTON_MIDDLE_PRESSED;
                xf86PostButtonEvent (pInfo->dev, TRUE, 2, 0, 0, 2, ax, ay);   /* MIDDLE released */
                TRACE(2, ErrorF("%s (%d, %d)  MIDDLE released\n", PM_INFO, ax, ay));
            }

            priv->hold_x = priv->old_x = 0;
            priv->hold_y = priv->old_y = 0;
            priv->touch_flags &= ~LOCK_WINDOW;

        }

        /*********************************************************
         * the untouch should always come after the button release
         *********************************************************/
        if ((priv->proximity == TRUE) && (priv->pen_down == 0)) {
            priv->proximity = FALSE;
            xf86PostProximityEvent(pInfo->dev, 0, 0, 2, ax, ay);
        }

    }/*for*/

}




static void ReadInputComPort (InputInfoPtr pInfo)
{
    PenMountPrivatePtr priv = (PenMountPrivatePtr) (pInfo->private);
    int x, y;
    int temp = 0;
    int pmode = 1;
    int ax, ay;
    unsigned char v1, v2;
    ScrnInfoPtr pScrn;
    Rotation rotation;

    TRACE(1, ErrorF("%s ReadInputComPort\n", PM_INFO));

    /* Get screen resolution again, beacase Ubuntu will change resolution after login X window */
    priv->screen_width  = screenInfo.screens[priv->screen_num]->width;
    priv->screen_height = screenInfo.screens[priv->screen_num]->height;

    /* 
     * set blocking to -1 on the first call because we know there is data to
     * read. Xisb automatically clears it after one successful read so that
     * succeeding reads are preceeded buy a select with a 0 timeout to prevent
     * read from blocking indefinately.
     */
    XisbBlockDuration (priv->buffer, -1);

    while (1) {


        if (PenMountGetPacket(priv) != Success)
            break;

        /* forced to 10-bit */
        if (priv->ad_bit == 10 && (priv->packet[2] > 0x03 || priv->packet[4] > 0x03))
            priv->ad_bit = 12;


        /*        [0]   [1] [2] [3] [4]
        /* 6000 : 70/30, XL, XH, YL, YH */
        /* 9000 : FF/BF, XH, XL, YH, YL (bit7 unused) */
        if (priv->ad_bit == 12 && priv->chip == PM6000COM) {
            /* convert to 10-bit */
                                         /* ----BA98 */
            v1 = priv->packet[2] & 0x03; /* ------98 */
            v1 <<= 6;                    /* 98------ */
            v1 &= 0xC0;
            v2 = priv->packet[1];        /* 76543210 */
            v2 >>= 2;                    /* --765432 */
            v2 &= 0x3F;
            v1 |= v2;                    /* 98765432 */
            priv->packet[1] = v1;
            
            v1 = priv->packet[2];        /* ----BA98 */
            v1 >>= 2;                    /* ------BA */
            v1 &= 0x03;
            priv->packet[2] = v1;

                                         /* ----BA98 */
            v1 = priv->packet[4] & 0x03; /* ------98 */
            v1 <<= 6;                    /* 98------ */
            v1 &= 0xC0;
            v2 = priv->packet[3];        /* 76543210 */
            v2 >>= 2;                    /* --765432 */
            v2 &= 0x3F;
            v1 |= v2;                    /* 98765432 */
            priv->packet[3] = v1;
            
            v1 = priv->packet[4];     /* ----BA98 */
            v1 >>= 2;                 /* ------BA */
            v1 &= 0x03;
            priv->packet[4] = v1;
            
        }

        switch(priv->chip) {
            case PM9000:
                priv->old_pendown = priv->pen_down;
                if (priv->packet[0] == 0xFF)
                    priv->pen_down = 1; /* pen down */
                else if (priv->packet[0] == 0xBF )
                    priv->pen_down = 0; /* pen up */

                /* 9000 : FF/BF, XH, XL, YH, YL (bit7 unused) */
                ax = priv->packet[1] & 0x07;        /* -----987 */
                ax <<= 7;                           /* ------98 7------- */
                ax += priv->packet[2];              /* ------98 76543210 */
                ax <<= 1;                           /* convert to 11-bit */

                ay = priv->packet[3] & 0x07;        /* -----987 */
                ay <<= 7;                           /* ------98 7------- */
                ay += priv->packet[4];              /* ------98 76543210 */
                ay <<= 1;                           /* convert to 11-bit */

                if (priv->swap_xy) {
                    y = ax;
                    x = ay;
                } else {
                    x = ax;
                    y = ay;
                }
                break;
            case PM6000COM:
                priv->old_pendown = priv->pen_down;
                /* none checksum */
                if ( priv->packet[0] == 0x70 )
                    priv->pen_down = 1; /* pen down */
                if ( priv->packet[0] == 0x30 )
                    priv->pen_down = 0; /* pen up */
     
                /* 6000 : 70/30, XL, XH, YL, YH */
                if (priv->ad_bit == 12) {
                    ax = ( priv->packet[2] & 0x0f ); /* ----BA98 */
                    ax <<= 8;                        /* ----BA98 -------- */
                    ax += priv->packet[1];           /* ----BA98 76543210 */
                    ax >>= 1;                        /* -----BA8 87654321 , 11-bit */

                    ay = ( priv->packet[4] & 0x0f ); /* ----BA98 */
                    ay <<= 8;                        /* ----BA98 -------- */
                    ay += priv->packet[3];           /* ----BA98 76543210 */
                    ay >>= 1;                        /* -----BA8 87654321 , 11-bit */
                } else {
                    ax = ( priv->packet[2] & 0x03 ); /* ------98 */
                    ax <<= 8;                        /* ------98 -------- */
                    ax += priv->packet[1];           /* ------98 76543210 */
                    ax <<= 1;                        /* -----987 6543210x , 11-bit */

                    ay = ( priv->packet[4] & 0x0f ); /* ------98 */
                    ay <<= 8;                        /* ------98 -------- */
                    ay += priv->packet[3];           /* ------98 76543210 */
                    ay <<= 1;                        /* -----987 6543210x , 11-bit */
                }

                if ( priv->swap_xy ) {
                    y = ax;
                    x = ay;
                } else {
                    x = ax;
                    y = ay;
                }
                
                TRACE(3, ErrorF("%s x=%d, y=%d\n", PM_INFO, x, y));
                break;
            default:
                break;
        }/*switch*/

        if (priv->old_pendown == 0 && priv->pen_down == 1) {    /* get pen-down moment */
            GET_TIME(tv0);
            
            priv->touch_flags = 0;  /* clear touch_flags */
            
            /* hold X, Y (raw) */
            priv->hold_x = x;
            priv->hold_y = y;

            priv->old_x  = x;
            priv->old_y  = y;
            priv->touch_flags |= LOCK_WINDOW;
            priv->touch_flags |= PEN_DOWN;
            
            PenMountBeep(priv, 1);
            
            priv->mcnt = 0;

            /* FIFO operation */
            IOWorkerThread(pInfo);

            g_last_packet_x = x;
            g_last_packet_y = y;

            TRACE(2, ErrorF("%s PEN-DOWN!!\n", PM_INFO));

            /******************  detect ROTATION *************/
            pScrn = xf86Screens[priv->screen_num];
            rotation = RRGetRotation(pScrn->pScreen);
            TRACE(2, ErrorF("PenMount : rotation = %d \n", rotation));
            if (rotation != priv->rotation) {
                priv->rotation = rotation;
                priv->CalibrationMode = SetCalibration(priv->configfile, priv->rotation);
                if (priv->CalibrationMode == 0)
                    LoadStdCalibration(priv);
            }
            g_last_rotation = rotation;

        } else if (priv->old_pendown == 1 && priv->pen_down == 0) { /* pen-up moment */
            priv->touch_flags |= PEN_UP;
            PenMountBeep(priv, 0);
            TRACE(2, ErrorF("%s PEN-UP!!\n", PM_INFO));
        }

        priv->old_x = x;
        priv->old_y = y;

        if (priv->CalibrationMode > 0) {
            /*************************************
             *       advanced calibration        *
             *************************************/
            TRACE(4, ErrorF("%s do_calib: raw(%d, %d)-->", PM_INFO, x, y));
            /* DO IT! */
            DoCalibrate(priv->xmin_offset, priv->xmax_offset, priv->ymin_offset, priv->ymax_offset, &x, &y);
            TRACE(4, ErrorF("%s do_advcal(%d, %d)-->", PM_INFO, x, y));

            if (priv->screen_scale == 1) {
                /* DON'T SCALED X, Y PLEASE..... */
                x = (x) * priv->screen_width  / (2047);
                y = (y) * priv->screen_height / (2047);
            }
            TRACE(4, ErrorF("%s scr(%d, %d), %dx%d\n", PM_INFO, x, y, priv->screen_width, priv->screen_height));
            
        } else {
            /*************************************
             *       standard calibration        *
             *************************************/

            /* Scaling */
            TRACE(4, ErrorF("%s do_stdcal: raw(%d, %d)-->", PM_INFO, x, y));

            if (priv->reporting_mode == TS_Scaled) {
                pmode = priv->pmode;
                
                /* swap x, y */
                if (pmode >= 5 && pmode <= 8) {
                    temp = x;
                    x    = y;
                    y    = temp;
                }

                if (pmode == 1 || pmode == 2 || pmode == 5 || pmode == 6) {
                    if (x < priv->std_xmin)
                        x = priv->std_xmin;
                    if (x > priv->std_xmax)
                        x = priv->std_xmax;
                    if (priv->screen_scale == 1)
                        x = ((x - priv->std_xmin) * priv->screen_width) / (priv->std_xmax - priv->std_xmin);
                    else
                        x = ((x - priv->std_xmin) * 2047) / (priv->std_xmax - priv->std_xmin);
                } else {
                    if (x < priv->std_xmin)
                        x = priv->std_xmin;
                    if (x > priv->std_xmax)
                        x = priv->std_xmax;
                    if (priv->screen_scale == 1)
                        x = ((priv->std_xmax - x) * priv->screen_width) / (priv->std_xmax - priv->std_xmin);
                    else
                        x = ((priv->std_xmax - x) * 2047) / (priv->std_xmax - priv->std_xmin);
                }

                if (pmode == 1 || pmode == 3 || pmode == 5 || pmode == 7) {
                    if (y < priv->std_ymin)
                        y = priv->std_ymin;
                    if (y > priv->std_ymax)
                        y = priv->std_ymax;
                    if (priv->screen_scale == 1)
                        y = ((y - priv->std_ymin) * priv->screen_height) / (priv->std_ymax - priv->std_ymin);
                    else
                        y = ((y - priv->std_ymin) * 2047) / (priv->std_ymax - priv->std_ymin);
                } else {
                    if (y < priv->std_ymin)
                        y = priv->std_ymin;
                    if (y > priv->std_ymax)
                        y = priv->std_ymax;
                    if (priv->screen_scale == 1)
                        y = ((priv->std_ymax - y) * priv->screen_height) / (priv->std_ymax - priv->std_ymin);
                    else
                        y = ((priv->std_ymax - y) * 2047) / (priv->std_ymax - priv->std_ymin);
                }
                TRACE(4, ErrorF("%s scr(%d, %d)\n", PM_INFO, x, y));
            }
        }/*end-if (CalibrationMode > 0)*/

        xf86XInputSetScreen(pInfo, priv->screen_num, x, y);

        /* 2008.05.13 added */
        if ((priv->proximity == FALSE) && (priv->pen_down == 1)) {
            priv->proximity = TRUE;
            xf86PostProximityEvent (pInfo->dev, 1, 0, 2, x, y);
        }

        /*
         * Send event
         * We *must* generate a motion before a button change if pointer
         * location has changed as DIX assumes this. This is why we always
         * emit a motion, regardless of the kind of packet processed.
         */
        switch(priv->average_point) {
            case 4:
                priv->av_x[priv->mcnt%4] = x;
                priv->av_y[priv->mcnt%4] = y;
                if (priv->mcnt < 4) {
                    ax = x;
                    ay = y;
                    xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, x, y);
                    TRACE(2, ErrorF("%s Motion[%d] : (%d, %d)\n", PM_INFO, priv->mcnt, x, y));
                } else {
                    ax = (priv->av_x[0]+priv->av_x[1]+priv->av_x[2]+priv->av_x[3])/4;
                    ay = (priv->av_y[0]+priv->av_y[1]+priv->av_y[2]+priv->av_y[3])/4;

                    /* slow down for PenMount 6000 USB */
                    if (priv->chip == PM6000USB) {
                        if (priv->mcnt%2)
                            xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, ax, ay);
                    } else {
                        xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, ax, ay);
                    }
                }
                break;
            case 8:
                priv->av_x[priv->mcnt%8] = x;
                priv->av_y[priv->mcnt%8] = y;
                if (priv->mcnt < 8) {
                    ax = x;
                    ay = y;
                    xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, x, y);
                    TRACE(2, ErrorF("%s Motion[%d] : (%d, %d)\n", PM_INFO, priv->mcnt, x, y));
                } else {
                    ax = (priv->av_x[0]+priv->av_x[1]+priv->av_x[2]+priv->av_x[3]+
                          priv->av_x[4]+priv->av_x[5]+priv->av_x[6]+priv->av_x[7])/8;
                    ay = (priv->av_y[0]+priv->av_y[1]+priv->av_y[2]+priv->av_y[3]+
                          priv->av_y[4]+priv->av_y[5]+priv->av_y[6]+priv->av_y[7])/8;
                    /* slow down for PenMount 6000 USB */
                    if (priv->chip == PM6000USB) {
                        if (priv->mcnt%2)
                            xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, ax, ay);
                    } else {
                        xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, ax, ay);
                    }
                }
                break;
            default:
                ax = x;
                ay = y;
                xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, ax, ay);
                break;
        }/* switch */
        priv->mcnt++;

        /********************/
        /*  Press and Hold  */
        /********************/
        if (priv->touch_flags & LOCK_WINDOW) {
            GET_TIME(tv1);

            temp = g_last_packet_x + priv->old_x;
            g_last_packet_x = temp / 2;

            temp = g_last_packet_y + priv->old_y;
            g_last_packet_y = temp / 2;


            if (priv->touch_flags & DOING_CALIB) {
                /* calib hold time > 2000ms */
                if (DIF_TIME(tv0, tv1) > priv->calib_hold_time) {
                    priv->touch_flags |= ISSUE_2ND_BUTTON;
                    priv->touch_flags &= ~LOCK_WINDOW;
                }
                /* exceed range */
                if (xf86abs(priv->hold_x - priv->old_x) >= priv->lock_window_range ||
                    xf86abs(priv->hold_y - priv->old_y) >= priv->lock_window_range)
                    priv->touch_flags &= ~LOCK_WINDOW;
            } else {
                /* hold time > 700ms */
                if (DIF_TIME(tv0, tv1) > priv->hold_time) {
                    priv->touch_flags |= ISSUE_2ND_BUTTON;
                    priv->touch_flags &= ~LOCK_WINDOW;
                }
                /* exceed range */
                if (xf86abs(priv->hold_x - priv->old_x) >= priv->lock_window_range ||
                    xf86abs(priv->hold_y - priv->old_y) >= priv->lock_window_range)
                    priv->touch_flags &= ~LOCK_WINDOW;
            }
        }

        /* xf86PostButtonEvent(device, is_absolute, button, is_down, first_valuator, num_valuator, valuator list....)
         * button : (1=left, 3=right)
         * is_down: (1=down, 0=up) */


        /****************/
        /*  2nd Button  */
        /****************/
        if (priv->touch_flags & ISSUE_2ND_BUTTON) {
            priv->touch_flags &= ~(ISSUE_2ND_BUTTON);

            createPenMountData();   /* prior to send post button event */

            /* Send Left released before send RIGHT/MIDDLE pressed */
            xf86PostButtonEvent (pInfo->dev, TRUE, 1, 0, 0, 2, ax, ay);   /* LEFT released */
            priv->touch_flags &= ~BUTTON_LEFT_PRESSED;

            if (priv->rightbutton == 1) {
                priv->touch_flags |= BUTTON_RIGHT_PRESSED;
                xf86PostButtonEvent (pInfo->dev, TRUE, 3, 1, 0, 2, ax, ay);   /* RIGHT pressed */
                TRACE(2, ErrorF("%s (%d, %d)  RIGHT pressed\n", PM_INFO, ax, ay));
            } else {
                priv->touch_flags |= BUTTON_MIDDLE_PRESSED;
                xf86PostButtonEvent (pInfo->dev, TRUE, 2, 1, 0, 2, ax, ay);   /* MIDDLE pressed */
                TRACE(2, ErrorF("%s (%d, %d)  MIDDLE pressed\n", PM_INFO, ax, ay));
            }

        }


        /*******************/
        /*    Pen Down     */
        /*******************/
        if (priv->touch_flags & PEN_DOWN) {
            priv->touch_flags &= ~PEN_DOWN;
            /* FIFO operation */
            /*IOWorkerThread(pInfo);*/
            /*createPenMountData();*/   /* prior to send post button event */

            #if 0
            if (xf86access("/tmp/penmount_calib", F_OK) == 0) {
                priv->hold_time_orig = priv->hold_time;
                priv->hold_time = 2000000;  /* 2 sec */
            }
            #endif


            if (priv->pendown_mode == 0) {
                /* stream mode */
                priv->touch_flags |= BUTTON_LEFT_PRESSED;
                xf86PostButtonEvent (pInfo->dev, TRUE, 1, 1, 0, 2, ax, ay);   /* LEFT pressed */
                TRACE(2, ErrorF("%s (%d, %d)  LEFT pressed\n", PM_INFO, ax, ay));
            } else {
                /* point mode */
                xf86PostButtonEvent (pInfo->dev, TRUE, 1, 1, 0, 2, ax, ay);   /* LEFT pressed */
                xf86PostButtonEvent (pInfo->dev, TRUE, 1, 0, 0, 2, ax, ay);   /* LEFT released*/
                TRACE(2, ErrorF("%s (%d, %d)  LEFT pressed-release\n", PM_INFO, ax, ay));
            }
            /* save pendown X, Y (screen) for something */
            priv->pd_x = x;
            priv->pd_y = y;
        }


        /******************/
        /*     Pen Up     */
        /******************/
        if (priv->touch_flags & PEN_UP) {
            /*createPenMountData();*/   /* prior to send post button event */
            priv->touch_flags &= ~PEN_UP;

            if (priv->touch_flags & BUTTON_LEFT_PRESSED) {
                /* LEFT pressed */
                priv->touch_flags &= ~BUTTON_LEFT_PRESSED;
                if (priv->touch_flags & LOCK_WINDOW) {
                    /* LEFT released*/
                    /*
                    xf86PostButtonEvent (pInfo->dev, TRUE, 1, 0, 0, 2, priv->pd_x, priv->pd_y);
                    TRACE(2, ErrorF("(%d, %d)  LEFT released (LOCK_WINDOW)\n", priv->pd_x, priv->pd_y));
                    */
                    xf86PostButtonEvent (pInfo->dev, TRUE, 1, 0, 0, 2, ax, ay);
                    TRACE(2, ErrorF("%s (%d, %d)  LEFT released (LOCK_WINDOW)\n", PM_INFO, ax, ay));

                } else {
                    xf86PostButtonEvent (pInfo->dev, TRUE, 1, 0, 0, 2, ax, ay);   /* LEFT released*/
                    TRACE(2, ErrorF("%s (%d, %d)  LEFT released (none LOCK_WINDOW)\n", PM_INFO, ax, ay));
                }
            }

            if (priv->touch_flags & BUTTON_RIGHT_PRESSED) {
                /* RIGHT pressed */
                priv->touch_flags &= ~BUTTON_RIGHT_PRESSED;
                xf86PostButtonEvent (pInfo->dev, TRUE, 3, 0, 0, 2, ax, ay);   /* RIGHT released */
                TRACE(2, ErrorF("%s (%d, %d)  RIGHT released\n", PM_INFO, ax, ay));
            } else if (priv->touch_flags & BUTTON_MIDDLE_PRESSED) {
                /* MIDDLE pressed */
                priv->touch_flags &= ~BUTTON_MIDDLE_PRESSED;
                xf86PostButtonEvent (pInfo->dev, TRUE, 2, 0, 0, 2, ax, ay);   /* MIDDLE released */
                TRACE(2, ErrorF("%s (%d, %d)  MIDDLE released\n", PM_INFO, ax, ay));
            }

            priv->hold_x = priv->old_x = 0;
            priv->hold_y = priv->old_y = 0;
            priv->touch_flags &= ~LOCK_WINDOW;

        }

        /*********************************************************
         * the untouch should always come after the button release
         *********************************************************/
        if ((priv->proximity == TRUE) && (priv->pen_down == 0)) {
            priv->proximity = FALSE;
            xf86PostProximityEvent(pInfo->dev, 0, 0, 2, ax, ay);
        }

    }/* while */

}


static void ReadInputLpc (InputInfoPtr pInfo)
{
    PenMountPrivatePtr priv = (PenMountPrivatePtr) (pInfo->private);
    int button = 0;
    int pmode = 1;
    int ax, ay;
    struct input_event  ev;
    int     len;
    int     x, y, temp;

    ScrnInfoPtr pScrn;
    Rotation rotation;


    TRACE(1, ErrorF("%s ReadInputLpc\n", PM_INFO));

    /* Get screen resolution again, beacase Ubuntu will change resolution after login X window */
    priv->screen_width  = screenInfo.screens[priv->screen_num]->width;
    priv->screen_height = screenInfo.screens[priv->screen_num]->height;

    /*TRACE(3, ErrorF("%s screen %d x %d\n", PM_INFO, priv->screen_width, priv->screen_height));*/
    TRACE(3, ErrorF("%s screen %d/%d/%d : %d x %d\n", PM_INFO, priv->screen_num, screenInfo.numScreens,
        screenInfo.numVideoScreens, priv->screen_width, priv->screen_height));

    /* unused, because of USB HID packet */
    /*priv->packet_size = 6;*/

    /*
    if (priv->old_x)
        x = (priv->old_x);
    if (priv->old_y)
        y = (priv->old_y);
    */

    while (xf86WaitForInput(pInfo->fd, 0) > 0) {

        TRACE(1, ErrorF("%s xf86WaitForInput\n", PM_INFO));
        len = read(pInfo->fd, &ev, sizeof(ev));
        if (len != sizeof(ev)) {
            TRACE(2, ErrorF("%s Read Error: len = %d <> %d\n", PM_ERROR, len, sizeof(ev)));
            if (len < 0) {
                xf86DisableDevice(pInfo->dev, TRUE);
                return;
            }
            break;
        }

        if (priv->old_x)
            x = (priv->old_x);
        if (priv->old_y)
            y = (priv->old_y);


        switch(ev.type) {
        case EV_ABS:
            if (ev.code == ABS_X) {
                x = ev.value;
                TRACE(2, ErrorF("%s EV_ABS-->ABS_X : %d => ", PM_INFO, x));
                x <<= 1;    /* 11-bit */
                TRACE(2, ErrorF("%d\n", x));
                priv->lex_mode |= 0x01;
            } else if (ev.code == ABS_Y) {
                y = ev.value;
                TRACE(2, ErrorF("%s EV_ABS-->ABS_Y : %d => ", PM_INFO, y));
                y <<= 1;    /* 11-bit */
                TRACE(2, ErrorF("%d\n", y));
                priv->lex_mode |= 0x02;
            }
            break;
        case EV_KEY:
            if (ev.code == BTN_TOUCH) {
                priv->old_pendown = priv->pen_down;
                priv->pen_down = ev.value;
                TRACE(2, ErrorF("%s EV_KEY-->BTN_TOUCH : %d\n", PM_INFO, priv->pen_down));
            }
            break;
        case EV_SYN:
            /*continue;*/
            break;
        }/* switch */

        if ((priv->lex_mode & 0x03) != 0x03) {
            if (x)
                priv->old_x = x;
            if (y)
                priv->old_y = y;
            TRACE(2, ErrorF("%s priv->lex_mode != 0x03,  %d", PM_INFO, priv->lex_mode));
            continue;
        }

        if (priv->old_pendown == 0 && priv->pen_down == 1) {    /* get pen-down moment */

            /* must update old_pendown here, because the pen-down info of packet just
               sent once when pen-down and sent pen-up info once when pen-up */
            priv->old_pendown = 1;
            
            GET_TIME(tv0);
            
            priv->touch_flags = 0;  /* clear touch_flags */
            
            /* hold X, Y (raw) */
            priv->hold_x = x;
            priv->hold_y = y;

            priv->old_x  = x;
            priv->old_y  = y;
            priv->touch_flags |= LOCK_WINDOW;
            priv->touch_flags |= PEN_DOWN;
            
            PenMountBeep(priv, 1);
            
            priv->mcnt = 0;

            /* FIFO operation */
            IOWorkerThread(pInfo);

            g_last_packet_x = x;
            g_last_packet_y = y;
            TRACE(2, ErrorF("%s PEN-DN : (%4d, %4d)\n", PM_INFO, x, y));
            
            /******************  detect ROTATION *************/
            pScrn = xf86Screens[priv->screen_num];
            rotation = RRGetRotation(pScrn->pScreen);
            TRACE(2, ErrorF("PenMount : rotation = %d \n", rotation));
            if (rotation != priv->rotation) {
                priv->rotation = rotation;
                priv->CalibrationMode = SetCalibration(priv->configfile, priv->rotation);
                if (priv->CalibrationMode == 0)
                    LoadStdCalibration(priv);
            }
            g_last_rotation = rotation;
            
        } else if (priv->old_pendown == 1 && priv->pen_down == 0) { /* pen-up moment */
            priv->touch_flags |= PEN_UP;
            PenMountBeep(priv, 0);
            TRACE(2, ErrorF("%s PEN-UP : (%4d, %4d)\n", PM_INFO, x, y));
            priv->lex_mode = 0;
            /* the reason as before mentioned */
            priv->old_pendown = 0;
        }

        priv->old_x = x;
        priv->old_y = y;

        if (priv->CalibrationMode > 0) {
            /*************************************
             *       advanced calibration        *
             *************************************/
            TRACE(2, ErrorF("%s Do_AdvCal : Raw(%4d, %4d)-->", PM_INFO, x, y));
            /* DO IT! */
            DoCalibrate(priv->xmin_offset, priv->xmax_offset, priv->ymin_offset, priv->ymax_offset, &x, &y);
            TRACE(2, ErrorF("Calib(%d, %d)-->", x, y));

            if (priv->screen_scale == 1) {
                /* DON'T SCALED X, Y PLEASE..... */
                /*
                x = (x) * priv->screen_width  / (1023);
                y = (y) * priv->screen_height / (1023);
                */
                x = (x) * priv->screen_width  / (2047);
                y = (y) * priv->screen_height / (2047);
            }
            TRACE(2, ErrorF("Scaled(%d, %d)  Res:%dx%d Sca:%d\n", x, y, priv->screen_width, priv->screen_height, priv->screen_scale));
            
        } else {
            /*************************************
             *       standard calibration        *
             *************************************/

            /* Scaling */
            TRACE(2, ErrorF("%s Do_StdCal : Raw(%4d, %4d)-->", PM_INFO, x, y));

            if (priv->reporting_mode == TS_Scaled) {
                pmode = priv->pmode;
                
                /* swap x, y */
                if (pmode >= 5 && pmode <= 8) {
                    temp = x;
                    x    = y;
                    y    = temp;
                }

                if (pmode == 1 || pmode == 2 || pmode == 5 || pmode == 6) {
                    if (x < priv->std_xmin)
                        x = priv->std_xmin;
                    if (x > priv->std_xmax)
                        x = priv->std_xmax;
                    if (priv->screen_scale == 1)
                        x = ((x - priv->std_xmin) * priv->screen_width) / (priv->std_xmax - priv->std_xmin);
                    else
                        /*x = ((x - priv->std_xmin) * 1023) / (priv->std_xmax - priv->std_xmin);*/
                        x = ((x - priv->std_xmin) * 2047) / (priv->std_xmax - priv->std_xmin);
                } else {
                    if (x < priv->std_xmin)
                        x = priv->std_xmin;
                    if (x > priv->std_xmax)
                        x = priv->std_xmax;
                    if (priv->screen_scale == 1)
                        x = ((priv->std_xmax - x) * priv->screen_width) / (priv->std_xmax - priv->std_xmin);
                    else
                        /*x = ((priv->std_xmax - x) * 1023) / (priv->std_xmax - priv->std_xmin);*/
                        x = ((priv->std_xmax - x) * 2047) / (priv->std_xmax - priv->std_xmin);
                }

                if (pmode == 1 || pmode == 3 || pmode == 5 || pmode == 7) {
                    if (y < priv->std_ymin)
                        y = priv->std_ymin;
                    if (y > priv->std_ymax)
                        y = priv->std_ymax;
                    if (priv->screen_scale == 1)
                        y = ((y - priv->std_ymin) * priv->screen_height) / (priv->std_ymax - priv->std_ymin);
                    else
                        /*y = ((y - priv->std_ymin) * 1023) / (priv->std_ymax - priv->std_ymin);*/
                        y = ((y - priv->std_ymin) * 2047) / (priv->std_ymax - priv->std_ymin);
                } else {
                    if (y < priv->std_ymin)
                        y = priv->std_ymin;
                    if (y > priv->std_ymax)
                        y = priv->std_ymax;
                    if (priv->screen_scale == 1)
                        y = ((priv->std_ymax - y) * priv->screen_height) / (priv->std_ymax - priv->std_ymin);
                    else
                        /*y = ((priv->std_ymax - y) * 1023) / (priv->std_ymax - priv->std_ymin);*/
                        y = ((priv->std_ymax - y) * 2047) / (priv->std_ymax - priv->std_ymin);
                }
                TRACE(2, ErrorF("Screen(%4d, %4d), Res:%dx%d, Sca:%d\n", x, y, priv->screen_width, priv->screen_height, priv->screen_scale));
            }
        }/*end-if (CalibrationMode > 0)*/

        /*
         * Send event
         * We *must* generate a motion before a button change if pointer
         * location has changed as DIX assumes this. This is why we always
         * emit a motion, regardless of the kind of packet processed.
         */
        switch(priv->average_point) {
            case 4:
                priv->av_x[priv->mcnt%4] = x;
                priv->av_y[priv->mcnt%4] = y;
                if (priv->mcnt < 4) {
                    ax = x;
                    ay = y;
                    /*xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, ax, ay);*/
                    TRACE(2, ErrorF("%s Motion4[%d] : (%d, %d)\n", PM_INFO, priv->mcnt, ax, ay));
                } else {
                    ax = (priv->av_x[0]+priv->av_x[1]+priv->av_x[2]+priv->av_x[3])/4;
                    ay = (priv->av_y[0]+priv->av_y[1]+priv->av_y[2]+priv->av_y[3])/4;
                    /*xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, ax, ay);*/
                    TRACE(2, ErrorF("%s Motion4[%d] : (%d, %d)\n", PM_INFO, priv->mcnt, ax, ay));
                }
                xf86XInputSetScreen(pInfo, priv->screen_num, ax, ay);
                if ((priv->proximity == FALSE) && (priv->pen_down == 1)) {
                    priv->proximity = TRUE;
                    xf86PostProximityEvent (pInfo->dev, 1, 0, 2, ax, ay);
                }
                xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, ax, ay);

                break;
            case 8:
                priv->av_x[priv->mcnt%8] = x;
                priv->av_y[priv->mcnt%8] = y;
                if (priv->mcnt < 8) {
                    ax = x;
                    ay = y;
                    /*xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, ax, ay);*/
                    TRACE(2, ErrorF("%s Motion8[%d] : (%d, %d)\n", PM_INFO, priv->mcnt, ax, ay));
                } else {
                    ax = (priv->av_x[0]+priv->av_x[1]+priv->av_x[2]+priv->av_x[3]+
                          priv->av_x[4]+priv->av_x[5]+priv->av_x[6]+priv->av_x[7])/8;
                    ay = (priv->av_y[0]+priv->av_y[1]+priv->av_y[2]+priv->av_y[3]+
                          priv->av_y[4]+priv->av_y[5]+priv->av_y[6]+priv->av_y[7])/8;
                    /*xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, ax, ay);*/
                    TRACE(2, ErrorF("%s Motion8[%d] : (%d, %d)\n", PM_INFO, priv->mcnt, ax, ay));
                }
                xf86XInputSetScreen(pInfo, priv->screen_num, ax, ay);
                if ((priv->proximity == FALSE) && (priv->pen_down == 1)) {
                    priv->proximity = TRUE;
                    xf86PostProximityEvent (pInfo->dev, 1, 0, 2, ax, ay);
                }
                xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, ax, ay);

                break;
            default:
                /*xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, x, y);*/
                ax = x;
                ay = y;
                TRACE(2, ErrorF("%s Motion0[%d] : (%d, %d)\n", PM_INFO, priv->mcnt, ax, ay));
                xf86XInputSetScreen(pInfo, priv->screen_num, ax, ay);
                if ((priv->proximity == FALSE) && (priv->pen_down == 1)) {
                    priv->proximity = TRUE;
                    xf86PostProximityEvent (pInfo->dev, 1, 0, 2, ax, ay);
                }
                xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, ax, ay);
                break;
        }/* switch */
        priv->mcnt++;


        /********************/
        /*  Press and Hold  */
        /********************/
        if (priv->touch_flags & LOCK_WINDOW) {
            GET_TIME(tv1);

            temp = g_last_packet_x + priv->old_x;
            g_last_packet_x = temp / 2;

            temp = g_last_packet_y + priv->old_y;
            g_last_packet_y = temp / 2;


            if (priv->touch_flags & DOING_CALIB) {
                /* calib hold time > 2000ms */
                if (DIF_TIME(tv0, tv1) > priv->calib_hold_time) {
                    priv->touch_flags |= ISSUE_2ND_BUTTON;
                    priv->touch_flags &= ~LOCK_WINDOW;
                }
                /* exceed range */
                if (xf86abs(priv->hold_x - priv->old_x) >= priv->lock_window_range ||
                    xf86abs(priv->hold_y - priv->old_y) >= priv->lock_window_range)
                    priv->touch_flags &= ~LOCK_WINDOW;
            } else {
                /* hold time > 700ms */
                if (DIF_TIME(tv0, tv1) > priv->hold_time) {
                    priv->touch_flags |= ISSUE_2ND_BUTTON;
                    priv->touch_flags &= ~LOCK_WINDOW;
                }
                /* exceed range */
                if (xf86abs(priv->hold_x - priv->old_x) >= priv->lock_window_range ||
                    xf86abs(priv->hold_y - priv->old_y) >= priv->lock_window_range)
                    priv->touch_flags &= ~LOCK_WINDOW;
            }
        }

        /* xf86PostButtonEvent(device, is_absolute, button, is_down, first_valuator, num_valuator, valuator list....)
         * button : (1=left, 3=right)
         * is_down: (1=down, 0=up) */


        /****************/
        /*  2nd Button  */
        /****************/
        if (priv->touch_flags & ISSUE_2ND_BUTTON) {
            priv->touch_flags &= ~(ISSUE_2ND_BUTTON);

            createPenMountData();   /* prior to send post button event */

            /* Send Left released before send RIGHT/MIDDLE pressed */
            xf86PostButtonEvent (pInfo->dev, TRUE, 1, 0, 0, 2, ax, ay);   /* LEFT released */
            priv->touch_flags &= ~BUTTON_LEFT_PRESSED;

            if (priv->rightbutton == 1) {
                priv->touch_flags |= BUTTON_RIGHT_PRESSED;
                xf86PostButtonEvent (pInfo->dev, TRUE, 3, 1, 0, 2, ax, ay);   /* RIGHT pressed */
                TRACE(2, ErrorF("%s (%d, %d)  RIGHT pressed\n", PM_INFO, ax, ay));
            } else {
                priv->touch_flags |= BUTTON_MIDDLE_PRESSED;
                xf86PostButtonEvent (pInfo->dev, TRUE, 2, 1, 0, 2, ax, ay);   /* MIDDLE pressed */
                TRACE(2, ErrorF("%s (%d, %d)  MIDDLE pressed\n", PM_INFO, ax, ay));
            }

        }


        /*******************/
        /*    Pen Down     */
        /*******************/
        if (priv->touch_flags & PEN_DOWN) {
            priv->touch_flags &= ~PEN_DOWN;
            /* FIFO operation */
            /*IOWorkerThread(pInfo);*/
            /*createPenMountData();*/   /* prior to send post button event */

            #if 0
            if (xf86access("/tmp/penmount_calib", F_OK) == 0) {
                priv->hold_time_orig = priv->hold_time;
                priv->hold_time = 2000000;  /* 2 sec */
            }
            #endif


            if (priv->pendown_mode == 0) {
                /* stream mode */
                priv->touch_flags |= BUTTON_LEFT_PRESSED;
                xf86PostButtonEvent (pInfo->dev, TRUE, 1, 1, 0, 2, ax, ay);   /* LEFT pressed */
                TRACE(2, ErrorF("%s (%d, %d)  LEFT pressed\n", PM_INFO, ax, ay));
            } else {
                /* point mode */
                xf86PostButtonEvent (pInfo->dev, TRUE, 1, 1, 0, 2, ax, ay);   /* LEFT pressed */
                xf86PostButtonEvent (pInfo->dev, TRUE, 1, 0, 0, 2, ax, ay);   /* LEFT released*/
                TRACE(2, ErrorF("%s (%d, %d)  LEFT pressed-release\n", PM_INFO, ax, ay));
            }
            /* save pendown X, Y (screen) for something */
            priv->pd_x = x;
            priv->pd_y = y;
        }


        /******************/
        /*     Pen Up     */
        /******************/
        if (priv->touch_flags & PEN_UP) {
            /*createPenMountData();*/   /* prior to send post button event */
            priv->touch_flags &= ~PEN_UP;

            if (priv->touch_flags & BUTTON_LEFT_PRESSED) {
                /* LEFT pressed */
                priv->touch_flags &= ~BUTTON_LEFT_PRESSED;
                if (priv->touch_flags & LOCK_WINDOW) {
                    /* LEFT released*/
                    /*
                    xf86PostButtonEvent (pInfo->dev, TRUE, 1, 0, 0, 2, priv->pd_x, priv->pd_y);
                    TRACE(2, ErrorF("(%d, %d)  LEFT released (LOCK_WINDOW)\n", priv->pd_x, priv->pd_y));
                    */
                    xf86PostButtonEvent (pInfo->dev, TRUE, 1, 0, 0, 2, ax, ay);
                    TRACE(2, ErrorF("%s (%d, %d)  LEFT released (LOCK_WINDOW)\n", PM_INFO, ax, ay));

                } else {
                    xf86PostButtonEvent (pInfo->dev, TRUE, 1, 0, 0, 2, ax, ay);   /* LEFT released*/
                    TRACE(2, ErrorF("%s (%d, %d)  LEFT released (none LOCK_WINDOW)\n", PM_INFO, ax, ay));
                }
            }

            if (priv->touch_flags & BUTTON_RIGHT_PRESSED) {
                /* RIGHT pressed */
                priv->touch_flags &= ~BUTTON_RIGHT_PRESSED;
                xf86PostButtonEvent (pInfo->dev, TRUE, 3, 0, 0, 2, ax, ay);   /* RIGHT released */
                TRACE(2, ErrorF("%s (%d, %d)  RIGHT released\n", PM_INFO, ax, ay));
            } else if (priv->touch_flags & BUTTON_MIDDLE_PRESSED) {
                /* MIDDLE pressed */
                priv->touch_flags &= ~BUTTON_MIDDLE_PRESSED;
                xf86PostButtonEvent (pInfo->dev, TRUE, 2, 0, 0, 2, ax, ay);   /* MIDDLE released */
                TRACE(2, ErrorF("%s (%d, %d)  MIDDLE released\n", PM_INFO, ax, ay));
            }

            priv->hold_x = priv->old_x = 0;
            priv->hold_y = priv->old_y = 0;
            priv->touch_flags &= ~LOCK_WINDOW;

        }

        /*********************************************************
         * the untouch should always come after the button release
         *********************************************************/
        if ((priv->proximity == TRUE) && (priv->pen_down == 0)) {
            priv->proximity = FALSE;
            xf86PostProximityEvent(pInfo->dev, 0, 0, 2, ax, ay);
        }

    }/* while */

}



/* The ControlProc function may need to be tailored for your device */
static int ControlProc (InputInfoPtr pInfo, xDeviceCtl * control)
{
    /*xDeviceTSCalibrationCtl *c = (xDeviceTSCalibrationCtl *) control;*/
    xDeviceAbsCalibCtl *c = (xDeviceAbsCalibCtl *) control;
    PenMountPrivatePtr priv = (PenMountPrivatePtr) (pInfo->private);

    /* debug */
    TRACE(1, ErrorF("%s ControlProc %d %d %d %d\n", PM_INFO, c->min_x, c->max_x, c->min_y, c->max_y));

    priv->min_x = c->min_x;
    priv->max_x = c->max_x;
    priv->min_y = c->min_y;
    priv->max_y = c->max_y;

    return (Success);
}


/* the CloseProc should not need to be tailored to your device */
static void CloseProc (InputInfoPtr pInfo)
{
    TRACE(1, ErrorF("%s CloseProc\n", PM_INFO));
}

/* The SwitchMode function may need to be tailored for your device */
static int SwitchMode (ClientPtr client, DeviceIntPtr dev, int mode)
{
    InputInfoPtr pInfo = dev->public.devicePrivate;
    PenMountPrivatePtr priv = (PenMountPrivatePtr) (pInfo->private);

    /* debug */
    TRACE(1, ErrorF("%s SwitchMode\n", PM_INFO));

    if ((mode == TS_Raw) || (mode == TS_Scaled)) {
        priv->reporting_mode = mode;
        return (Success);
    } else if ((mode == SendCoreEvents) || (mode == DontSendCoreEvents)) {
        /*xf86XInputSetSendCoreEvents (pInfo, (mode == SendCoreEvents));*/
        return (Success);
    } else {
        return (!Success);
    }
}


/*
 * The ConvertProc function may need to be tailored for your device.
 * This function converts the device's valuator outputs to x and y coordinates
 * to simulate mouse events.
 */
static Bool ConvertProc (InputInfoPtr pInfo,
             int first,
             int num,
             int v0,
             int v1,
             int v2,
             int v3,
             int v4,
             int v5,
             int *x,
             int *y)
{
    PenMountPrivatePtr priv = (PenMountPrivatePtr) (pInfo->private);
    ScrnInfoPtr pScrn = xf86Screens[priv->screen_num];
    Rotation rotation = RRGetRotation(pScrn->pScreen);
    int     tmp0, tmp1;


    TRACE(1, ErrorF("%s ConvertProc\n", PM_INFO));

    if (priv->reporting_mode == TS_Raw) {

        *x = xf86ScaleAxis (v0, 0, priv->screen_width, priv->min_x, priv->max_x);
        *y = xf86ScaleAxis (v1, 0, priv->screen_height, priv->min_y, priv->max_y);

        TRACE(4, ErrorF("%s in ConvertProc TS_RAW. \n", PM_INFO));

    } else {
/*
        switch(rotation) {
        case RR_Rotate_0:
            break;
        case RR_Rotate_180:
            tmp0 = 2047 - v0;
            tmp1 = 2047 - v1;
            v0 = tmp0;
            v1 = tmp1;
            break;
        case RR_Rotate_90:
            tmp1 = v0;
            tmp0 = 2047 - v1;
            v0 = tmp0;
            v1 = tmp1;
            break;
        case RR_Rotate_270:
            tmp1 = 2047 - v0;
            tmp0 = v1;
            v0 = tmp0;
            v1 = tmp1;
            break;
        default:
            break;
		}
*/
        *x = v0;
        *y = v1;
    }

    return (TRUE);
}

/*
 * standard calibration /etc/penmount.dat
 
 * <CalibPointsNum>  0
 * <Top Point>       64 64
 * <Right Point>     1984 64
 * <Bottom Point>    1984 1984
 * <Left Point>      64 1984
 * <Rotate>          1|2|4|8
 */
static void LoadStdCalibration(PenMountPrivatePtr priv)
{
    XF86FILE *f = xf86fopen(priv->configfile, "r");
    int     px[4];
    int     py[4];
    int     i;
    int     rotate;
    int     tmpx[4], tmpy[4];
    
    int     rotate_table[4][8] = {
            { 1, 2, 3, 4, 5, 6, 7, 8 },  /* normal   */
            { 7, 5, 8, 6, 3, 1, 4, 2 },  /* left     */
            { 4, 3, 2, 1, 8, 7, 6, 5 },  /* inverted */
            { 6, 8, 5, 7, 2, 4, 1, 3 }}; /* right    */

    if(!f) {
        /* avoid bad value cause the system crash */    
        priv->std_xmin = 20;
        priv->std_ymin = 20;
        priv->std_xmax = 2000;
        priv->std_ymax = 2000;
        priv->pmode = 1;
        return;
    }
    /*
         +--0------------+
         |               |
         |               1
         |               |
         3               |
         |               |
         +------------2--+
     */

    xf86fscanf(f, "%d", &i);  /* useless */
    xf86fscanf(f, "%d", &px[0]);
    xf86fscanf(f, "%d", &py[0]);
    xf86fscanf(f, "%d", &px[1]);
    xf86fscanf(f, "%d", &py[1]);
    xf86fscanf(f, "%d", &px[2]);
    xf86fscanf(f, "%d", &py[2]);
    xf86fscanf(f, "%d", &px[3]);
    xf86fscanf(f, "%d", &py[3]);
    /*f86fscanf(f, "%d", &priv->pmode);*/
    xf86fscanf(f, "%d", &rotate);
    xf86fclose(f);

    for (i=0; i<4; i++) {
        tmpx[i] = px[i];
        tmpy[i] = py[i];
    }    
    /* rotation */
    switch(rotate) {
    case 1: /* normal */
        break;
    case 2: /* left */
        px[0] = tmpx[1];
        py[0] = tmpy[1];
        px[1] = tmpx[2];
        py[1] = tmpy[2];
        px[2] = tmpx[3];
        py[2] = tmpy[3];
        px[3] = tmpx[0];
        py[3] = tmpy[0];
        break;
    case 4: /* inverted */
        px[0] = tmpx[2];
        py[0] = tmpy[2];
        px[1] = tmpx[3];
        py[1] = tmpy[3];
        px[2] = tmpx[0];
        py[2] = tmpy[0];
        px[3] = tmpx[1];
        py[3] = tmpy[1];
        break;
    case 8: /* right */
        px[0] = tmpx[3];
        py[0] = tmpy[3];
        px[1] = tmpx[0];
        py[1] = tmpy[0];
        px[2] = tmpx[1];
        py[2] = tmpy[1];
        px[3] = tmpx[2];
        py[3] = tmpy[2];
        break;
    }

    /* determine portrait mode */
    if ( (px[3]<px[0] && px[0]<px[2] && px[2]<px[1]) &&
         (py[0]<py[1] && py[1]<py[3] && py[3]<py[2]) )
        priv->pmode = 1;
    else if ( (px[3]<px[0] && px[0]<px[2] && px[2]<px[1]) &&
              (py[2]<py[3] && py[3]<py[1] && py[1]<py[0]) )
        priv->pmode = 2;
    else if ( (px[1]<px[2] && px[2]<px[0] && px[0]<px[3]) &&
              (py[0]<py[1] && py[1]<py[3] && py[3]<py[2]) )
        priv->pmode = 3;
    else if ( (px[1]<px[2] && px[2]<px[0] && px[0]<px[3]) &&
              (py[2]<py[3] && py[3]<py[1] && py[1]<py[0]) )
        priv->pmode = 4;
    else if ( (px[0]<px[1] && px[1]<px[3] && px[3]<px[2]) &&
              (py[3]<py[0] && py[0]<py[2] && py[2]<py[1]) )
        priv->pmode = 5;
    else if ( (px[2]<px[3] && px[3]<px[1] && px[1]<px[0]) &&
              (py[3]<py[0] && py[0]<py[2] && py[2]<py[1]) )
        priv->pmode = 6;
    else if ( (px[0]<px[1] && px[1]<px[3] && px[3]<px[2]) &&
              (py[1]<py[2] && py[2]<py[0] && py[0]<py[3]) )
        priv->pmode = 7;
    else if ( (px[2]<px[3] && px[3]<px[1] && px[1]<px[0]) &&
              (py[1]<py[2] && py[2]<py[0] && py[0]<py[3]) )
        priv->pmode = 8;
    else
        priv->pmode = 1;


    priv->std_xmin = priv->std_ymin = 65535;
    priv->std_xmax = priv->std_ymax = 0;
    for (i=0; i<4; i++) {
        if (px[i] > priv->std_xmax)
            priv->std_xmax = px[i];
        if (px[i] < priv->std_xmin)
            priv->std_xmin = px[i];
        if (py[i] > priv->std_ymax)
            priv->std_ymax = py[i];
        if (py[i] < priv->std_ymin)
            priv->std_ymin = py[i];
    }

/*
    if (priv->std_xmin > 511)
        priv->std_xmin = 20;
    if (priv->std_ymin > 511)
        priv->std_ymin = 20;
    if (priv->std_xmax < 1535)
        priv->std_xmax = 2000;
    if (priv->std_ymax < 1535)
        priv->std_ymax = 2000;
*/
  
    TRACE(3, ErrorF("%s Load calibrate : %d, %d, %d, %d, %d\n", PM_INFO,
        priv->pmode, priv->std_xmin, priv->std_xmax, priv->std_ymin, priv->std_ymax));

    if (priv->rotation == 1)
        rotate = 0;
    else if (priv->rotation == 2)
        rotate = 1;
    else if (priv->rotation == 4)
        rotate = 2;
    else if (priv->rotation == 8)
        rotate = 3;
    else
        rotate = 0;

    i = (priv->pmode - 1);    /* original pmode 1~8 --> 0~7 */
    priv->pmode = rotate_table[rotate][i];

    TRACE(2, ErrorF("%s pmode : %d, rotate_table[%d][%d]\n", PM_INFO,
        priv->pmode, rotate, i));
    
}


static void
PenMountPtrCtrl (DeviceIntPtr device, PtrCtrl *ctrl)
{
  /* I have no clue what this does, except that registering it stops the 
     X server segfaulting in ProcGetPointerMapping()
     Ho Hum.
  */
    TRACE(1, ErrorF("%s PenMountPtrCtrl\n", PM_INFO));

}
/*
static void beep_beep(void)
{
    XF86FILE    *f;

    f = xf86fopen("/dev/console", "w");
    xf86fprintf(f, "\x07");
    xf86fclose(f);
}
*/
static int sys_beep(int freq, int vol, int dur)
{
    int     fd;
    int     f;
    
    if (freq < 20)
        f = 20;
    else if (freq > 32767)
        f = 32767;
    else
        f = freq;
    
    fd = open("/dev/console", O_WRONLY);
    if (!fd) {
        ErrorF("%s Cann't open /dev/console for beeper\n", PM_ERROR);
        return -1;
    }

    ioctl(fd, KIOCSOUND, (int)(1193180/f));
    usleep(vol*dur*20);
    ioctl(fd, KIOCSOUND, 0);
    
    close(fd);
    
}

/* cannot call xf86SoundKbdBell @ XINPUT_ABI_VERSIOM <> 0 */    
/*xf86SoundKbdBell(priv->press_vol, priv->press_pitch, priv->press.dur);*/

static void PenMountBeep (PenMountPrivatePtr priv, int press)
{
    TRACE(1, ErrorF("%s PenMountBeep\n", PM_INFO));

    if (priv->beep) {
        if (press) {
            sys_beep(priv->press_pitch, priv->press_vol, priv->press_dur);
        } else {
            sys_beep(priv->rel_pitch, priv->rel_vol, priv->rel_dur);
        }
    }
}


static Bool PenMountGetPacket (PenMountPrivatePtr priv)
{
    int     cnt;
    int     c;

    TRACE(1, ErrorF("%s PenMountGetPacket\n", PM_INFO));
    
    cnt = 0;

    while ((c = XisbRead(priv->buffer)) >= 0) {

        /* xf86Msg( X_PROBED, "PenMountGetPacket = %02x\n", c); */

        if (cnt++ > 500)
            return !Success;

        switch (priv->lex_mode) {
            case PM_Byte0:
                if (priv->chip == PM9000) {
                    if ( c != 0xFF && c != 0xBF )
                        return (!Success);
                } else if (priv->chip == PM6000COM ) {
                    if ( c != 0x70 && c != 0x30 )
                        return (!Success);
                }
                priv->packet[0] = (unsigned char) c;
                priv->lex_mode  = PM_Byte1;
                break;
            case PM_Byte1:
                priv->packet[1] = (unsigned char) c;
                priv->lex_mode  = PM_Byte2;
                break;
            case PM_Byte2:
                priv->packet[2] = (unsigned char) c;
                priv->lex_mode  = PM_Byte3;
                break;
            case PM_Byte3:
                priv->packet[3] = (unsigned char) c;
                priv->lex_mode  = PM_Byte4;
                break;
            case PM_Byte4:
                priv->packet[4] = (unsigned char) c;
                if (priv->chip == PM9000) {
                    priv->lex_mode  = PM_Byte0;
                    return Success;
                } else if (priv->chip == PM6000COM) {
                    priv->lex_mode  = PM_Byte5;
                }
                break;
            case PM_Byte5:
                priv->packet[5] = (unsigned char) c;
                priv->lex_mode  = PM_Byte0;
                return Success;
                break;
            default:
                break;
        }
    }

    return !Success;
}

static void PenMountUnInit(InputDriverPtr drv, InputInfoRec *pInfo, int flags)
{
    PenMountPrivatePtr priv = (PenMountPrivatePtr) (pInfo->private);

    xf86DeleteInput(pInfo, 0);

}

#if 0
================================================================================

static Bool PenMountSendPacket(PenMountPrivatePtr priv, unsigned char *buf, int len)
{
    int count = 0;

    TRACE(1, ErrorF("%s PenMountSendPacket %02x %02x %02x %02x %02x\n", PM_INFO,
          buf[0], buf[1], buf[2], buf[3], buf[4]));

    while(len > 0) {
        if (XisbWrite(priv->buffer, buf, 1) == 1) {
            buf++;
            len--;
            continue;
        }
        if (count++ > 500)
            break;
    }
    if (len > 0) {
        TRACE(3, ErrorF("%s PenMountSendPacket send fail, len=%d\n", PM_INFO, len));
    } else {
        TRACE(3, ErrorF("%s PenMountSendPacket send success\n", PM_INFO));
    }
    
    return (len ? !Success : Success);

}

static Bool PenMountGetProduct(PenMountPrivatePtr priv)
{
    int count = 0;
    int c;

    TRACE(1, ErrorF("%s PenMountGetProduct\n", PM_INFO));

    while ((c = XisbRead(priv->buffer)) >= 0) {
        /*
		 * fail after 500 bytes so the server doesn't hang forever if a
		 * device sends bad data.
		 */
        if (count++ > 500)
            return (!Success);

        switch(priv->lex_mode) {
            case PM_Byte0:
            case PM_Byte1:
            case PM_Byte2:
            case PM_Byte3:
            case PM_Byte4:
            case PM_Byte5:
                break;
            case PM_Res0:
                if ( c == 0xF1 )
                    priv->lex_mode = PM_Res1;
                priv->packet[0] = (unsigned char) c;
                TRACE(3, ErrorF("%s PenMountGetProduct : packet[0] = %02x\n", PM_INFO, priv->packet[0]));
                break;
            case PM_Res1:
                priv->packet[1] = (unsigned char) c;
                priv->lex_mode = PM_Res2;
                break;
            case PM_Res2:
                priv->packet[2] = (unsigned char) c;
                priv->lex_mode = PM_Res3;
                break;
            case PM_Res3:
                priv->packet[3] = (unsigned char) c;
                priv->lex_mode = PM_Res4;
                break;
            case PM_Res4:
                priv->packet[4] = (unsigned char) c;
                priv->lex_mode = PM_Res5;
                break;
            case PM_Res5:
                priv->packet[5] = (unsigned char) c;
                priv->lex_mode = PM_Byte0;
                TRACE(3, ErrorF("%s PenMountGetProduct: %02x %02x %02x %02x %02x %02x\n", 
                    PM_INFO, priv->packet[0], priv->packet[1], priv->packet[2],
                    priv->packet[3], priv->packet[4],  priv->packet[5]));
                return (Success);
                break;
        }/* switch */
        
    }/* while */
    return (!Success);
}

static Bool PenMountGetVersion(PenMountPrivatePtr priv)
{
    int count = 0;
    int c;

    TRACE(1, ErrorF("%s PenMountGetVersion\n", PM_INFO));

    while ((c = XisbRead(priv->buffer)) >= 0) {
        /*
		 * fail after 500 bytes so the server doesn't hang forever if a
		 * device sends bad data.
		 */
        if (count++ > 500)
            return (!Success);

        switch(priv->lex_mode) {
            case PM_Byte0:
            case PM_Byte1:
            case PM_Byte2:
            case PM_Byte3:
            case PM_Byte4:
                break;
            case PM_Res0:
                if ( c == 0xf2 )
                    priv->lex_mode = PM_Res1;
                priv->packet[0] = (unsigned char) c;
                TRACE(3, ErrorF("%s PenMountGetVersion : packet[0] = %02x\n", PM_INFO, priv->packet[0]));
                break;
            case PM_Res1:
                if ( c == 0xd9 )
                    priv->lex_mode = PM_Res2;
                priv->packet[1] = (unsigned char) c;
                break;
            case PM_Res2:
                priv->packet[2] = (unsigned char) c;
                priv->lex_mode = PM_Res3;
                break;
            case PM_Res3:
                priv->packet[3] = (unsigned char) c;
                priv->lex_mode = PM_Res4;
                break;
            case PM_Res4:
                priv->packet[4] = (unsigned char) c;
                priv->lex_mode = PM_Byte0;
                TRACE(3, ErrorF("%s PenMountGetVersion: %02x %02x %02x %02x %02x\n", 
                    PM_INFO, priv->packet[0], priv->packet[1], priv->packet[2],
                    priv->packet[3], priv->packet[4]));
                return (Success);
                break;
        }/* switch */
        
    }/* while */
    return (!Success);
}
================================================================================
#endif

#if 0
================================================================================
static Bool QueryHardware(InputInfoPtr pInfo)
{
    PenMountPrivatePtr priv = (PenMountPrivatePtr) (pInfo->private);
    unsigned char       product_6000_cmd[6] = { 0xee, 0, 0, 0, 0, 0x11 };
    unsigned char       search_9000_cmd[5] = { 0xf2, 0, 0, 0, 0 };

    TRACE(1, ErrorF("%s QueryHardware\n", PM_INFO));

    if (PenMountSendPacket(priv, product_6000_cmd, 6) != Success) {
        TRACE(1, ErrorF("%s PenMountSendPacket failed!\n", PM_ERROR));
    }
    
    priv->lex_mode = PM_Res0;
    if (PenMountGetProduct(priv) == Success) {
        TRACE(1, ErrorF("%s Detect PM6000 OK!\n", PM_INFO));
        priv->chip = PM6000COM;
        return Success;
    }
    /* check 6000 failed, then check 9000 */
    if (priv->chip != PM6000COM) {
        xf86FlushInput(pInfo->fd);
        if (PenMountSendPacket(priv, search_9000_cmd, 5) != Success) {
            TRACE(1, ErrorF("%s PenMountSendPacket failed!\n", PM_ERROR));
        }
        if (PenMountGetVersion(priv) == Success) {
            TRACE(1, ErrorF("%s Detect PM9000 OK!\n", PM_INFO));
            priv->chip = PM9000;
            return Success;
        }
    }

    return (!Success);

}
================================================================================
#endif
