/* 
 * PenMount
 */

#ifndef _PENMOUNT_H_
#define _PENMOUNT_H_

#include <linux/input.h>
/**********************************************
          File Path
 **********************************************/
/*#define HALF_DUPLEX_READ    "/tmp/penmount_cmd"*/ /* pipe */
/*#define HALF_DUPLEX_WRITE   "/tmp/penmount_dat"*/ /* pipe */
/*#define CALIBRATION_DATA    "/etc/penmount.dat"*/ /* data */
#define CALIBRATION_DATA_FILE    "/penmount.dat" /* data */

/******************************************************************************
 *        Definitions
 *****************************************************************************/
#define PM_BUFFER_SIZE    1024
#define PM_MAX_BUTTONS    3       /* num of buttons */

#define PM_DEF_PMODE      1
/*
#define PM_DEF_MIN_X      0
#define PM_DEF_MAX_X      9500
#define PM_DEF_MIN_Y      0
#define PM_DEF_MAX_Y      10500
*/


/* log @ /var/log/Xorg.0.log */
#define X_PENMOUNT        " PenMount:"
const char *PM_INFO       = {"(II)" X_PENMOUNT}; /* Informational message */
const char *PM_ERROR      = {"(EE)" X_PENMOUNT}; /* Error message */



/* priv->touch_flags */
#define DOING_CALIB             0x80
#define BUTTON_RIGHT_PRESSED    0x40
#define BUTTON_MIDDLE_PRESSED   0x20
#define BUTTON_LEFT_PRESSED     0x10
#define LOCK_WINDOW             0x08
#define ISSUE_2ND_BUTTON        0x04
#define PEN_UP                  0x02
#define PEN_DOWN                0x01

typedef enum {

    CHIP_UNKNOWN = 0,
    PM9000,
    PM5000,
    PM6000USB,
    PM6000COM,
    PMLPC

} PenMountChip;


typedef enum {

    PM_Byte0,
    PM_Byte1,
    PM_Byte2, 
    PM_Byte3, 
    PM_Byte4,
    PM_Byte5,
    PM_Res0,
    PM_Res1,
    PM_Res2,
    PM_Res3,
    PM_Res4,
    PM_Res5,

} PenMountState;


/******************************************************************************
 *        Private Record
 *****************************************************************************/
typedef struct _PenMountPrivateRec {

    int     CalibrationMode;    /* 0:std_calib, >0:adv_calib, <0:error */

    int     pmode;              /* portrait mode, =1~8, for std_calib */
    int     min_x;              /* Minimum x reported by calibration */
    int     max_x;              /* Maximum x                         */
    int     min_y;              /* Minimum y reported by calibration */
    int     max_y;              /* Maximum y                         */
    int     screen_num;         /* screen no. of X window, Screen 
                                   associated with the device */
    int     screen_width;       /* Width of screen resolution X      */
    int     screen_height;      /* Height of screen resolution Y     */
    int     swap_xy;            /* swap X, Y */
    Bool    button_down;        /* is the "button" currently down    */
    
    int     old_x;              /* Prior x                           */
    int     old_y;              /* Prior y                           */
    int     old_button;         /* Prior button                      */
    
    int     hold_x;             /* pen-down moment X (raw) */
    int     hold_y;             /* pen-down moment Y (raw) */
    int     pd_x;               /* pen-down moment X (screen) */
    int     pd_y;               /* pen-down moment Y (screen) */

    int     proximity;

    char    pen_down;           /* 1=pen-down now, 0=pen-up now */
    char    old_pendown;        /* prior pen_down */

    int     button_number;      /* which button to report            */
    int     reporting_mode;     /* TS_Raw or TS_Scaled               */


    XISBuffer   *buffer;        /* raw data buffer */
    struct input_event ev;      /* packet being/just read */

    int     ad_bit;             /* A/D bit, 10=10bit, 12=12bit */
    int     pendown_mode;       /* 0:stream mode, 1=point mode */

    int     beep;               /* 0 = no beep, 1 = beep enabled          */
    int     press_vol;          /* volume of beep (press event)           */
    int     press_pitch;        /* pitch of beep (press event)            */
    int     press_dur;          /* length of beep in 10ms (press event)   */
    int     rel_vol;            /* volume of beep (release event)         */
    int     rel_pitch;          /* pitch of beep (release event)          */
    int     rel_dur;            /* length of beep in 10ms (release event) */

    int     rightbutton;        /* 1=right button, 0=middle button */

    int     hold_time;          /* press press n hold time to send button */
    int     calib_hold_time;    /* hold_time @ calibration */

    int     chip;               /* which model of PenMount chip */
    
    unsigned short   touch_flags;/* PenMount stat */
    
    int     mcnt;               /* counter for count num of packet */
    int     av_x[8];            /* 4 points average X (screen) */
    int     av_y[8];            /* 4 points average X (screen) */
    
    int     evfd;               /* event device file descriptor */
    char    evnode[32];         /* event device name */
    
    char    screen_scale;       /* screen resolution scale */
    char    average_point;      /* average 4/8 point for sent x, y (0=none, 4=4 point, 8=8 point*/
    int     lock_window_range;  /* press N hold range (raw data) */
    
    int     xmin_offset;        /* edge compensation */
    int     xmax_offset;
    int     ymin_offset;
    int     ymax_offset;
    int     std_xmin;
    int     std_xmax;
    int     std_ymin;
    int     std_ymax;
    char    configfile[64];
    int     lex_mode;
    unsigned char packet[6];
    int     packet_size;
    char    press_and_hold;
    char    rotation;
}
PenMountPrivateRec, *PenMountPrivatePtr;

/******************************************************************************
 *        Declarations
 *****************************************************************************/
static InputInfoPtr PenMountPreInit(InputDriverPtr drv, IDevPtr dev, int flags);
static void PenMountUnInit(InputDriverPtr drv, InputInfoRec *pInfo, int flags);

static Bool DeviceControl (DeviceIntPtr, int);
static void ReadInput (InputInfoPtr);
static void ReadInputComPort (InputInfoPtr);
static void ReadInputLpc (InputInfoPtr);
static int  ControlProc (InputInfoPtr, xDeviceCtl *);
static void CloseProc (InputInfoPtr);
static int  SwitchMode (ClientPtr, DeviceIntPtr, int);
static Bool ConvertProc (InputInfoPtr, int, int, int, int, int, int, int, int, int *, int *);

static void PenMountPtrCtrl(DeviceIntPtr device, PtrCtrl *ctrl);
static void LoadStdCalibration(PenMountPrivatePtr priv);
static void PenMountBeep(PenMountPrivatePtr priv, int press);
static Bool PenMountGetPacket (PenMountPrivatePtr priv);
/*
static Bool PenMountSendPacket(PenMountPrivatePtr priv, unsigned char *buf, int len);
static Bool PenMountGetProduct(PenMountPrivatePtr priv);
static Bool PenMountGetVersion(PenMountPrivatePtr priv);
static Bool QueryHardware(InputInfoPtr pInfo);
*/
#endif /* _PENMOUNT_H_ */
